""" Module to decouple cg code from Housekeeper code """
import datetime as dt
import logging
import os
from pathlib import Path
from typing import Iterable, List, Optional, Set, Tuple

from alchy import Query
from housekeeper.include import checksum as hk_checksum
from housekeeper.include import include_version
from housekeeper.store import Store, models
from housekeeper.store.models import File, Version

from cg.exc import HousekeeperVersionMissingError

LOG = logging.getLogger(__name__)


class HousekeeperAPI:
    """API to decouple cg code from Housekeeper"""

    def __init__(self, config: dict) -> None:
        self._store = Store(config["housekeeper"]["database"], config["housekeeper"]["root"])
        self.root_dir = config["housekeeper"]["root"]

    def __getattr__(self, name):
        LOG.warning("Called undefined %s on %s, please wrap", name, self.__class__.__name__)
        return getattr(self._store, name)

    def new_bundle(self, name: str, created_at: dt.datetime = None) -> models.Bundle:
        """Create a new file bundle"""
        return self._store.new_bundle(name, created_at)

    def add_bundle(self, bundle_data) -> Tuple[models.Bundle, models.Version]:
        """Build a new bundle version of files"""
        return self._store.add_bundle(bundle_data)

    def bundle(self, name: str) -> models.Bundle:
        """Fetch a bundle"""
        return self._store.bundle(name)

    def bundles(self) -> List[models.Bundle]:
        """Fetch bundles"""
        return self._store.bundles()

    def create_new_bundle_and_version(self, name: str) -> models.Bundle:
        """Create new bundle with version"""
        new_bundle: models.Bundle = self.new_bundle(name=name)
        self.add_commit(new_bundle)
        new_version: models.Version = self.new_version(created_at=new_bundle.created_at)
        new_bundle.versions.append(new_version)
        self.commit()
        LOG.info("New bundle created with name %s", new_bundle.name)
        return new_bundle

    def set_to_archive(self, file: models.File, value: bool) -> None:
        """Sets the 'to_archive' field of a file"""
        file.to_archive = value
        self.commit()

    def new_file(
        self, path: str, checksum: str = None, to_archive: bool = False, tags: list = None
    ) -> models.File:
        """Create a new file"""
        if tags is None:
            tags = []
        return self._store.new_file(path, checksum, to_archive, tags)

    def get_file(self, file_id: int) -> Optional[models.File]:
        """Fetch a file based on file id"""
        LOG.info("Fetching file %s", file_id)
        file_obj = self._store.file_(file_id)
        if not file_obj:
            LOG.info("file not found")
            return None
        return file_obj

    def delete_file(self, file_id: int) -> Optional[models.File]:
        """Delete a file both from database and disk (if included)"""
        file_obj: models.File = self.get_file(file_id)
        if not file_obj:
            LOG.info("Could not find file %s", file_id)
            return

        if file_obj.is_included and Path(file_obj.full_path).exists():
            LOG.info("Deleting file %s from disc", file_obj.full_path)
            Path(file_obj.full_path).unlink()

        LOG.info("Deleting file %s from housekeeper", file_id)
        file_obj.delete()
        self._store.commit()

        return file_obj

    def delete_file_if_related(self, stem: str, hk_file: File):
        """Delete a file if the full path includes the root"""
        if stem in hk_file.path:
            self.delete_file(file_id=hk_file.id)
            self._store.commit()
            LOG.info(f"HousekeeperAPI: {hk_file.path} deleted from Housekeeper")

    def check_for_files(self, bundle: str = None, tags=None, version=None) -> bool:
        """Check if there are files for a bundle, tags, and/or version"""
        files: Optional[Iterable[File]] = self.files(bundle=bundle, tags=tags, version=version)
        if not any(files):
            return False
        return True

    def add_file(
        self, path, version_obj: models.Version, tags: list, to_archive: bool = False
    ) -> models.File:
        """Add a file to the database"""
        if isinstance(tags, str):
            tags = [tags]
        for tag_name in tags:
            if not self.tag(tag_name):
                self.add_tag(tag_name)

        new_file = self.new_file(
            path=str(Path(path).absolute()),
            to_archive=to_archive,
            tags=[self.tag(tag_name) for tag_name in tags],
        )

        new_file.version = version_obj
        return new_file

    def files(
        self,
        *,
        bundle: str = None,
        tags: List[str] = None,
        version: int = None,
        path: str = None,
    ) -> Query:
        """Fetch files"""
        return self._store.files(bundle=bundle, tags=tags, version=version, path=path)

    @staticmethod
    def fetch_file_from_version(
        version_obj: models.Version, tags: Set[str]
    ) -> Optional[models.File]:
        """Fetch file that includes at least all tags in 'tags'

        Return None if no file could be found
        """
        LOG.debug("Fetch files from version with tags %s", tags)
        for file_obj in version_obj.files:
            tag: models.Tag
            file_tags = {tag.name for tag in file_obj.tags}
            if tags.issubset(file_tags):
                LOG.debug("Found file %s", file_obj)
                return file_obj
        LOG.info("Could not find any files matching the tags")

    def rollback(self):
        """Wrap method in Housekeeper Store"""
        return self._store.rollback()

    def session_no_autoflush(self):
        """Wrap property in Housekeeper Store"""
        return self._store.session.no_autoflush

    def get_files(
        self, bundle: str, tags: Optional[list] = None, version: Optional[int] = None
    ) -> Query:
        """Fetch all the files in housekeeper, optionally filtered by bundle and/or tags and/or
        version

        Returns:
            iterable(hk.Models.File)
        """
        return self._store.files(bundle=bundle, tags=tags, version=version)

    def check_bundle_files(
        self,
        bundle_name: str,
        file_paths: List[Path],
        last_version: Version,
        tags: Optional[list] = None,
    ) -> List[Path]:
        """Checks if any of the files in the provided list are already added to the provided
        bundle. Returns a list of files that have not been added"""
        for file in self.get_files(bundle=bundle_name, tags=tags, version=last_version.id):
            if Path(file.path) in file_paths:
                file_paths.remove(Path(file.path))
                LOG.info(
                    f"Path {file.path} is already linked to bundle {bundle_name} in housekeeper"
                )
        return file_paths

    @staticmethod
    def get_included_path(
        root_dir: Path, version_obj: models.Version, file_obj: models.File
    ) -> Path:
        """Generate the path to a file that should be included.
        If the version dir does not exist, create a new version dir in root dir
        """
        # generate root directory
        version_root_dir = root_dir / version_obj.relative_root_dir
        version_root_dir.mkdir(parents=True, exist_ok=True)
        LOG.info("Created new bundle version dir: %s", version_root_dir)
        return version_root_dir / Path(file_obj.path).name

    def include_file(self, file_obj: models.File, version_obj: models.Version) -> models.File:
        """Call the include version function to import related assets."""
        global_root_dir = Path(self.get_root_dir())

        new_path = self.get_included_path(global_root_dir, version_obj, file_obj)
        if file_obj.to_archive:
            # calculate sha1 checksum if file is to be archived
            file_obj.checksum = HousekeeperAPI.checksum(file_obj.path)
        # hardlink file to the internal structure
        os.link(file_obj.path, new_path)
        LOG.info("Linked file: %s -> %s", file_obj.path, new_path)
        file_obj.path = str(new_path).replace(f"{global_root_dir}/", "", 1)
        return file_obj

    def new_version(
        self, created_at: dt.datetime, expires_at: dt.datetime = None
    ) -> models.Version:
        """Create a new bundle version"""
        return self._store.new_version(created_at, expires_at)

    def version(self, bundle: str, date: dt.datetime) -> models.Version:
        """Fetch a version"""
        LOG.info("Fetch version %s from bundle %s", date, bundle)
        return self._store.version(bundle, date)

    def last_version(self, bundle: str) -> models.Version:
        """Gets the latest version of a bundle"""
        LOG.info("Fetch latest version from bundle %s", bundle)
        return (
            self._store.Version.query.join(models.Version.bundle)
            .filter(models.Bundle.name == bundle)
            .order_by(models.Version.created_at.desc())
            .first()
        )

    def get_create_version(self, bundle: str) -> models.Version:
        """Returns the latest version of a bundle if it exists. If not creates a bundle and
        returns its version"""
        last_version: models.Version = self.last_version(bundle=bundle)
        if not last_version:
            LOG.info("Creating bundle for sample %s in housekeeper", bundle)
            bundle_result: Tuple[models.Bundle, models.Version] = self.add_bundle(
                bundle_data={
                    "name": bundle,
                    "created_at": dt.datetime.now(),
                    "expires_at": None,
                    "files": [],
                }
            )
            last_version: models.Version = bundle_result[1]
        return last_version

    def new_tag(self, name: str, category: str = None):
        """Create a new tag"""
        return self._store.new_tag(name, category)

    def add_tag(self, name: str, category: str = None) -> models.Tag:
        """Add a tag to the database"""
        tag_obj = self._store.new_tag(name, category)
        self.add_commit(tag_obj)
        return tag_obj

    def tag(self, name: str) -> models.Tag:
        """Fetch a tag"""
        return self._store.tag(name)

    @staticmethod
    def get_tag_names_from_file(file: models.File) -> List[str]:
        """Fetch tag names for a file"""
        return [tag.name for tag in file.tags]

    def include(self, version_obj: models.Version):
        """Call the include version function to import related assets."""
        include_version(self.get_root_dir(), version_obj)
        version_obj.included_at = dt.datetime.now()

    def add_commit(self, *args, **kwargs):
        """Wrap method in Housekeeper Store"""
        return self._store.add_commit(*args, **kwargs)

    def commit(self):
        """Wrap method in Housekeeper Store"""
        return self._store.commit()

    def get_root_dir(self):
        """Returns the root dir of Housekeeper"""
        return self.root_dir

    @staticmethod
    def checksum(path):
        """Calculate the checksum"""
        return hk_checksum(path)

    def initialise_db(self):
        """Create all tables in the store."""
        self._store.create_all()

    def destroy_db(self):
        """Drop all tables in the store"""
        self._store.drop_all()

    def add_and_include_file_to_latest_version(self, case_id: str, file: Path, tags: list) -> None:
        version_obj: Version = self.last_version(case_id)
        if not version_obj:
            LOG.info("Family ID: %s not found in housekeeper", case_id)
            raise HousekeeperVersionMissingError
        file_obj = self.add_file(version_obj=version_obj, tags=tags, path=str(file.absolute()))
        self.include_file(version_obj=version_obj, file_obj=file_obj)
        self.commit()

    def find_file_in_latest_version(self, case_id: str, tags: list) -> Optional[File]:
        version_obj: Version = self.last_version(case_id)
        if not version_obj:
            LOG.info("Family ID: %s not found in housekeeper", case_id)
            raise HousekeeperVersionMissingError
        file: File = self.files(version=version_obj.id, tags=tags).first()
        return file



"""Report field validation helper"""

from typing import Optional

from cg.constants import NA_FIELD
from cg.models.report.report import ReportModel


def get_million_read_pairs(reads: int) -> Optional[float]:
    """Represents the number of sequencing reads as millions of read pairs."""

    return round(reads / 2_000_000, 1) if reads or isinstance(reads, int) else None


def get_missing_fields(empty_fields: list, required_fields: list) -> list:
    """Extracts the missing fields that are required to generate successfully the delivery report."""

    missing_fields = list()

    for field in empty_fields:
        if field in required_fields:
            missing_fields.append(field)

    return missing_fields


def get_empty_fields(report_data: dict) -> list:
    """Returns a list of empty report fields."""

    empty_fields = list()

    for field, value in report_data.items():
        if not value or value == NA_FIELD:
            if isinstance(value, bool):
                continue

            empty_fields.append(field)

    return empty_fields


def get_empty_report_data(report_data: ReportModel) -> dict:
    """Retrieve empty fields from a report data model."""

    empty_fields = {
        "report": get_empty_fields(report_data.dict()),
        "customer": get_empty_fields(report_data.customer.dict()),
        "case": get_empty_fields(report_data.case.dict()),
        "applications": {
            app.tag: get_empty_fields(app.dict())
            for app in report_data.case.applications
            if get_empty_fields(app.dict())
        },
        "data_analysis": get_empty_fields(report_data.case.data_analysis.dict()),
        "samples": {
            sample.id: get_empty_fields(sample.dict())
            for sample in report_data.case.samples
            if get_empty_fields(sample.dict())
        },
        "methods": {
            sample.id: get_empty_fields(sample.methods.dict())
            for sample in report_data.case.samples
            if get_empty_fields(sample.methods.dict())
        },
        "timestamps": {
            sample.id: get_empty_fields(sample.timestamps.dict())
            for sample in report_data.case.samples
            if get_empty_fields(sample.timestamps.dict())
        },
        "metadata": {
            sample.id: get_empty_fields(sample.metadata.dict())
            for sample in report_data.case.samples
            if get_empty_fields(sample.metadata.dict())
        },
    }

    # Clear empty values
    empty_fields = {k: v for k, v in empty_fields.items() if v}

    return empty_fields


def get_missing_report_data(empty_fields: dict, required_fields: dict) -> dict:
    """Retrieve the missing required fields from a report data model."""

    nested_sources = ["applications", "samples", "methods", "timestamps", "metadata"]
    missing_fields = dict()

    for source in empty_fields:
        if source in nested_sources:
            # Associates application/sample tags/ids to missing fields
            missing_data = {
                tag: get_missing_fields(empty_fields[source][tag], required_fields[source][tag])
                for tag in empty_fields[source]
                if get_missing_fields(empty_fields[source][tag], required_fields[source][tag])
            }
        else:
            missing_data = get_missing_fields(empty_fields[source], required_fields[source])

        if missing_data:
            missing_fields.update({source: missing_data})

    return missing_fields



"""Constants for delivery."""

from cgmodels.cg.constants import Pipeline

ONLY_ONE_CASE_PER_TICKET = [
    Pipeline.FASTQ,
    Pipeline.MICROSALT,
    Pipeline.SARS_COV_2,
]

SKIP_MISSING = [
    Pipeline.FASTQ,
    Pipeline.MICROSALT,
    Pipeline.SARS_COV_2,
]

BALSAMIC_ANALYSIS_CASE_TAGS = [
    {"delivery-report"},
    {"multiqc-html"},
    {"metrics"},
    {"visualization"},
    {"coverage"},
    {"vcf-sv-clinical"},
    {"vcf-sv-clinical-index"},
    {"vcf-sv-research"},
    {"vcf-sv-research-index"},
    {"vcf-snv-clinical"},
    {"vcf-snv-clinical-index"},
    {"vcf-snv-research"},
    {"vcf-snv-research-index"},
    {"germline", "vcf"},
    {"germline", "vcf-index"},
    {"vcf2cytosure"},
]

BALSAMIC_ANALYSIS_SAMPLE_TAGS = [
    {"cram"},
    {"cram-index"},
]

BALSAMIC_QC_ANALYSIS_CASE_TAGS = [
    {"multiqc-html"},
]

BALSAMIC_QC_ANALYSIS_SAMPLE_TAGS = [
    {"qc-cram"},
    {"qc-cram-index"},
]

BALSAMIC_UMI_ANALYSIS_CASE_TAGS = [
    {"vcf-umi-snv-clinical"},
    {"vcf-umi-snv-clinical-index"},
    {"vcf-umi-snv-research"},
    {"vcf-umi-snv-research-index"},
]

BALSAMIC_UMI_ANALYSIS_CASE_TAGS.extend(BALSAMIC_ANALYSIS_CASE_TAGS)

BALSAMIC_UMI_ANALYSIS_SAMPLE_TAGS = [
    {"umi-cram"},
    {"umi-cram-index"},
]

BALSAMIC_UMI_ANALYSIS_SAMPLE_TAGS.extend(BALSAMIC_ANALYSIS_SAMPLE_TAGS)


MIP_DNA_ANALYSIS_CASE_TAGS = [
    {"delivery-report"},
    {"vcf-clinical-sv-bin"},
    {"vcf-clinical-sv-bin-index"},
    {"vcf-research-sv-bin"},
    {"vcf-research-sv-bin-index"},
    {"gbcf"},
    {"gbcf-index"},
    {"snv-gbcf"},
    {"snv-gbcf-index"},
    {"snv-bcf"},
    {"snv-bcf-index"},
    {"sv-bcf"},
    {"sv-bcf-index"},
    {"vcf-snv-clinical"},
    {"vcf-snv-clinical-index"},
    {"vcf-snv-research"},
    {"vcf-snv-research-index"},
    {"vcf-sv-clinical"},
    {"vcf-sv-clinical-index"},
    {"vcf-sv-research"},
    {"vcf-sv-research-index"},
]

MIP_DNA_ANALYSIS_SAMPLE_TAGS = [{"bam"}, {"bam-index"}, {"cram"}, {"cram-index"}]

MIP_RNA_ANALYSIS_CASE_TAGS = [
    {"fusion", "clinical", "pdf"},
    {"fusion", "research", "pdf"},
    {"fusion", "vcf"},
    {"fusion", "vcf-index"},
    {"vcf-snv-clinical"},
    {"vcf-snv-clinical-index"},
    {"vcf-snv-research"},
    {"vcf-snv-research-index"},
]

MIP_RNA_ANALYSIS_SAMPLE_TAGS = [
    {"fusion", "star-fusion"},
    {"fusion", "arriba"},
    {"cram"},
    {"cram-index"},
    {"fusion", "vcf"},
    {"fusion", "vcf-index"},
    {"salmon-quant"},
]

MICROSALT_ANALYSIS_CASE_TAGS = [
    {"microsalt-qc"},
    {"microsalt-type"},
    {"assembly"},
    {"trimmed-forward-reads"},
    {"trimmed-reverse-reads"},
    {"reference-alignment-deduplicated"},
]

MICROSALT_ANALYSIS_SAMPLE_TAGS = []

FASTQ_ANALYSIS_CASE_TAGS = []

FASTQ_ANALYSIS_SAMPLE_TAGS = [
    {"fastq"},
]

SARSCOV2_ANALYSIS_CASE_TAGS = [
    {"pangolin"},
    {"ks-delivery"},
]

SARSCOV2_ANALYSIS_SAMPLE_TAGS = [
    {"fastq"},
]

PIPELINE_ANALYSIS_TAG_MAP = {
    Pipeline.BALSAMIC: {
        "case_tags": BALSAMIC_ANALYSIS_CASE_TAGS,
        "sample_tags": BALSAMIC_ANALYSIS_SAMPLE_TAGS,
    },
    Pipeline.BALSAMIC_QC: {
        "case_tags": BALSAMIC_QC_ANALYSIS_CASE_TAGS,
        "sample_tags": BALSAMIC_QC_ANALYSIS_SAMPLE_TAGS,
    },
    Pipeline.BALSAMIC_UMI: {
        "case_tags": BALSAMIC_UMI_ANALYSIS_CASE_TAGS,
        "sample_tags": BALSAMIC_UMI_ANALYSIS_SAMPLE_TAGS,
    },
    Pipeline.MIP_DNA: {
        "case_tags": MIP_DNA_ANALYSIS_CASE_TAGS,
        "sample_tags": MIP_DNA_ANALYSIS_SAMPLE_TAGS,
    },
    Pipeline.MIP_RNA: {
        "case_tags": MIP_RNA_ANALYSIS_CASE_TAGS,
        "sample_tags": MIP_RNA_ANALYSIS_SAMPLE_TAGS,
    },
    Pipeline.MICROSALT: {
        "case_tags": MICROSALT_ANALYSIS_CASE_TAGS,
        "sample_tags": MICROSALT_ANALYSIS_SAMPLE_TAGS,
    },
    Pipeline.FASTQ: {
        "case_tags": FASTQ_ANALYSIS_CASE_TAGS,
        "sample_tags": FASTQ_ANALYSIS_SAMPLE_TAGS,
    },
    Pipeline.SARS_COV_2: {
        "case_tags": SARSCOV2_ANALYSIS_CASE_TAGS,
        "sample_tags": SARSCOV2_ANALYSIS_SAMPLE_TAGS,
    },
}

PIPELINE_ANALYSIS_OPTIONS = PIPELINE_ANALYSIS_TAG_MAP.keys()



"""
    Module for loqusdb API
"""
import logging
from pathlib import Path
from subprocess import CalledProcessError

from cg.constants.constants import FileFormat
from cg.exc import CaseNotFoundError, DeleteCaseError
from cg.io.controller import ReadStream
from cg.utils import Process

LOG = logging.getLogger(__name__)


class LoqusdbAPI:

    """
    API for loqusdb
    """

    def __init__(self, config: dict, analysis_type: str = "wgs"):
        super(LoqusdbAPI, self).__init__()

        self.analysis_type = analysis_type

        self.loqusdb_config = config["loqusdb"]["config_path"]
        self.loqusdb_binary = config["loqusdb"]["binary_path"]

        if self.analysis_type == "wes":
            self.loqusdb_config = config["loqusdb_wes"]["config_path"]
            self.loqusdb_binary = config["loqusdb_wes"]["binary_path"]

        self.process = Process(binary=self.loqusdb_binary, config=self.loqusdb_config)

    def load(
        self,
        family_id: str,
        ped_path: Path,
        vcf_path: Path,
        gbcf_path: Path,
        vcf_sv_path: Path = None,
    ) -> dict:
        """Add observations from a VCF."""
        load_call_parameters = [
            "load",
            "-c",
            family_id,
            "-f",
            ped_path.as_posix(),
            "--variant-file",
            vcf_path.as_posix(),
            "--check-profile",
            gbcf_path.as_posix(),
            "--hard-threshold",
            "0.95",
            "--soft-threshold",
            "0.90",
            "--gq-treshold",
            "10",
        ]
        if self.analysis_type == "wgs" and vcf_sv_path:
            load_call_parameters.extend(["--sv-variants", vcf_sv_path.as_posix()])

        nr_variants = 0
        self.process.run_command(parameters=load_call_parameters)
        for line in self.process.stderr_lines():
            line_content = line.split("INFO")[-1].strip()
            if "inserted" in line_content:
                nr_variants = int(line_content.split(":")[-1].strip())

        return dict(variants=nr_variants)

    def delete_case(self, case_id: str) -> None:
        """Remove a case from LoqusDB."""
        delete_call_parameters = ["delete", "-c", case_id]

        self.process.run_command(parameters=delete_call_parameters)
        for line in self.process.stderr_lines():
            if f"INFO Removing case {case_id}" in line:
                return None
            if f"WARNING Case {case_id} does not exist" in line:
                raise CaseNotFoundError(f"Case {case_id} not found in loqusdb")

        # This should not happen. If it does, other exit messages must be handle
        raise DeleteCaseError(f"Could not delete case {case_id}")

    def get_case(self, case_id: str) -> dict:
        """Find a case in the database by case id."""
        cases_parameters = ["cases", "-c", case_id, "--to-json"]

        self.process.run_command(parameters=cases_parameters)

        # If case not in loqusdb, stdout of loqusdb command will be empty.
        if not self.process.stdout:
            raise CaseNotFoundError(f"Case {case_id} not found in loqusdb")

        return ReadStream.get_content_from_stream(
            file_format=FileFormat.JSON, stream=self.process.stdout
        )[0]

    def get_duplicate(self, vcf_file: Path) -> dict:
        """Find matching profiles in loqusdb"""
        duplicates_params = [
            "profile",
            "--check-vcf",
            vcf_file.as_posix(),
            "--profile-threshold",
            "0.95",
        ]

        try:
            self.process.run_command(parameters=duplicates_params)
        except CalledProcessError:
            # If CalledProcessError is raised, log and raise error
            LOG.critical("Could not run profile command")
            raise

        if not self.process.stdout:
            LOG.info("No duplicates found")
            return {}

        return ReadStream.get_content_from_stream(
            file_format=FileFormat.JSON, stream=self.process.stdout
        )

    def case_exists(self, case_id: str) -> bool:
        """Check if a case exists in LoqusDB"""
        try:
            self.get_case(case_id)
            return True
        except CaseNotFoundError:
            return False

    def __repr__(self):

        return f"LoqusdbAPI(binary={self.loqusdb_binary}," f"config={self.loqusdb_config})"



from cg.constants.nipt import Q30_THRESHOLD
from cg.meta.upload.nipt import NiptUploadAPI
from cg.models.cg_config import CGConfig


def test_flowcell_passed_qc_value(case_id: str, nipt_upload_api_context: CGConfig):

    # GIVEN a NiptUploadAPI context with a case that has been sequenced
    nipt_upload_api: NiptUploadAPI = NiptUploadAPI(nipt_upload_api_context)

    # WHEN checking the Q30 and read values on the flow cell
    flowcell_qc_value: bool = nipt_upload_api.flowcell_passed_qc_value(
        case_id=case_id, q30_threshold=Q30_THRESHOLD
    )

    # THEN the successful flow cell shall pass and
    assert flowcell_qc_value


def test_flowcell_failed_qc_value(
    nipt_upload_api_failed_fc_context: CGConfig, case_id: str, sample_id: str
):

    # GIVEN a NiptUploadAPI context with a case that has been sequenced, however the case is lacking reads
    nipt_upload_api: NiptUploadAPI = NiptUploadAPI(nipt_upload_api_failed_fc_context)

    # WHEN checking the Q30 and read values on the flowcell
    flowcell_qc_pass: bool = nipt_upload_api.flowcell_passed_qc_value(
        case_id=case_id, q30_threshold=Q30_THRESHOLD
    )

    # THEN the failed flow cell should have not been passed
    assert not flowcell_qc_pass



from cg.store import Store


def test_microbial_sample_to_dict(microbial_store: Store, helpers):

    # GIVEN a store with a Microbial sample
    sample_obj = helpers.add_microbial_sample(microbial_store)
    microbial_store.add_commit(sample_obj)
    assert sample_obj

    # WHEN running to dict on that sample
    a_dict = sample_obj.to_dict(links=True)

    # THEN you should get a dictionary with
    assert a_dict["id"]
    assert a_dict["internal_id"]
    assert a_dict["name"]
    assert a_dict["application_version_id"]
    assert a_dict["created_at"]
    assert a_dict["priority"]
    assert a_dict["reads"]
    assert a_dict["comment"]
    assert a_dict["application"]
    assert a_dict["application_version"]


def test_no_collaborators(base_store):
    # GIVEN a customer without collaborations
    new_customer_id = "cust004"
    customer = base_store.add_customer(
        new_customer_id,
        "No-colab",
        scout_access=True,
        invoice_address="Test street",
        invoice_reference="ABCDEF",
    )
    # WHEN calling the collaborators property
    collaborators = customer.collaborators
    # THEN only the customer should be returned
    assert len(collaborators) == 1
    assert collaborators.pop().internal_id == new_customer_id


def test_collaborators(base_store, customer_id):
    # GIVEN a customer with one collaboration
    customer = base_store.customer(customer_id)
    assert all(
        customer_obj.internal_id
        in [
            "cust001",
            "cust002",
            "cust003",
            customer_id,
        ]
        for customer_obj in customer.collaborations[0].customers
    )

    # WHEN calling the collaborators property
    # THEN the customer and the collaborators should be returned
    collaborators = customer.collaborators
    assert len(collaborators) == 4
    assert all(
        collaborator.internal_id
        in [
            "cust001",
            "cust002",
            "cust003",
            customer_id,
        ]
        for collaborator in collaborators
    )


def test_multiple_collaborations(base_store, customer_id):
    # Given a customer in two collaborations
    collaboration = base_store.add_collaboration("small_collaboration", "small collaboration")
    new_customer_id = "cust004"
    new_customer = base_store.add_customer(
        new_customer_id,
        "No-colab",
        scout_access=True,
        invoice_address="Test street",
        invoice_reference="ABCDEF",
    )
    prod_customer = base_store.customer(customer_id)
    collaboration.customers.extend([prod_customer, new_customer])
    base_store.add_commit(new_customer, collaboration)
    base_store.refresh(collaboration)
    # WHEN calling the collaborators property
    collaborators = prod_customer.collaborators
    # THEN all customers in both collaborations should be returned
    assert len(collaborators) == 5
    assert all(
        collaborator.internal_id in ["cust001", new_customer_id, "cust002", "cust003", customer_id]
        for collaborator in collaborators
    )



"""Tests for cleaning flow cell run directories using
cg.meta.clean.flow_cell_run_directories.RunDirFlowCell """
from unittest import mock

from cg.constants import HousekeeperTags
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.meta.clean.flow_cell_run_directories import RunDirFlowCell


@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_age(
    mock_statusdb,
    mock_hk,
    flowcell_path,
    timestamp_yesterday,
):
    # GIVEN a flow cell with a sequenced date:
    flow_cell_obj = RunDirFlowCell(flowcell_path, mock_statusdb, mock_hk)
    flow_cell_obj._sequenced_date = timestamp_yesterday

    # WHEN determining the age of a flow cell
    result = flow_cell_obj.age

    # THEN the age property of the flow cell should be set to 1 day
    assert result == 1


@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_sequenced_date_from_statusdb(
    mock_statusdb,
    mock_hk,
    flowcell_path,
    timestamp_yesterday,
):
    # GIVEN a flow cell with a sequenced_at date in statusdb
    flow_cell_obj = RunDirFlowCell(flowcell_path, mock_statusdb, mock_hk)
    mock_statusdb.flowcell.return_value.sequenced_at = timestamp_yesterday

    # WHEN determining the age of a flow cell
    result = flow_cell_obj.sequenced_date

    # THEN the sequenced_date property of the flow cell should be set to the sequenced_at date in
    # statusdb
    assert result == timestamp_yesterday


@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_sequenced_date_from_run_name(
    mock_statusdb,
    mock_hk,
    flowcell_path,
):
    # GIVEN a flow cell that does not exist in statusdb
    flow_cell_obj = RunDirFlowCell(flowcell_path, mock_statusdb, mock_hk)
    mock_statusdb.flowcell.return_value = None

    # WHEN determining the age of a flow cell
    result = flow_cell_obj.sequenced_date

    # THEN the sequenced_date property of the flow cell should be derived from the run name of
    # the flow cell
    assert result == flow_cell_obj.derived_date


@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_archive_sample_sheet_no_bundle(mock_statusdb, mock_hk, flowcell_path, novaseq_dir):
    # GIVEN a flow cell
    flow_cell_obj = RunDirFlowCell(flowcell_path, mock_statusdb, mock_hk)
    # GIVEN a sample sheet connected to the flow cell
    flow_cell_obj.sample_sheet_path = (
        novaseq_dir / DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
    )
    # GIVEN there is no bundle for the flow cell
    mock_hk.bundle.return_value = None
    # GIVEN the sample sheet does not exist in Housekeeper
    mock_hk.files.return_value.first.return_value = None
    mock_hk.get_files.return_value.first.return_value.is_included = False

    # WHEN archiving the sample sheet
    flow_cell_obj.archive_sample_sheet()

    # THEN a bundle and version should be created for the flow cell
    flow_cell_obj.hk.create_new_bundle_and_version.assert_called_once_with(name=flow_cell_obj.id)

    # THEN the sample sheet should be be added to Housekeeper
    flow_cell_obj.hk.add_and_include_file_to_latest_version.assert_called_once_with(
        case_id=flow_cell_obj.id,
        file=flow_cell_obj.sample_sheet_path,
        tags=[HousekeeperTags.ARCHIVED_SAMPLE_SHEET, flow_cell_obj.id],
    )



"""Tests for BALSAMIC analysis"""
from pathlib import Path

import pytest

from cg.constants.subject import Gender
from cg.exc import BalsamicStartError

from cg.meta.workflow.balsamic import BalsamicAnalysisAPI


def test_get_verified_gender():
    """Tests gender extraction from a sample dictionary"""

    # GIVEN a sample object
    sample_obj = {
        "ACC0000A0": {"gender": "female"},
        "ACC0000A1": {"gender": "female"},
    }

    # WHEN extracting the gender
    retrieved_gender = BalsamicAnalysisAPI.get_verified_gender(sample_obj)

    # THEN gender must match the expected one
    assert retrieved_gender == "female"


def test_get_verified_gender_error():
    """Tests gender extraction from a sample dictionary when two different gender has been provided"""

    # GIVEN a sample object with different genders
    sample_obj = {
        "ACC0000A0": {"gender": "male"},
        "ACC0000A1": {"gender": "female"},
    }

    # WHEN extracting the gender
    with pytest.raises(BalsamicStartError):
        BalsamicAnalysisAPI.get_verified_gender(sample_obj)


def test_get_verified_gender_unknown(caplog):
    """Tests gender extraction from a sample dictionary when the gender is unknown"""

    # GIVEN a sample object with different genders
    sample_obj = {
        "ACC0000A0": {"gender": "unknown"},
        "ACC0000A1": {"gender": "unknown"},
    }

    # WHEN extracting the gender
    retrieved_gender = BalsamicAnalysisAPI.get_verified_gender(sample_obj)

    # THEN gender must match the expected one
    assert retrieved_gender == Gender.FEMALE
    assert f"The provided gender is unknown, setting {Gender.FEMALE} as the default" in caplog.text


def test_get_verified_pon():
    """Tests PON verification"""

    # GIVEN specific panel bed and PON files
    panel_bed = "/path/gmcksolid_4.1_hg19_design.bed"
    pon_cnn = "/path/PON/gmcksolid_4.1_hg19_design_CNVkit_PON_reference_v2.cnn"
    invalid_pon_cnn = "/path/PON/gmssolid_15.2_hg19_design_CNVkit_PON_reference_v2.cnn"

    # WHEN validating the PON
    validated_pon = BalsamicAnalysisAPI.get_verified_pon(None, panel_bed, pon_cnn)

    # THEN the PON verification should be performed successfully
    assert pon_cnn == validated_pon
    with pytest.raises(BalsamicStartError):
        BalsamicAnalysisAPI.get_verified_pon(None, panel_bed, invalid_pon_cnn)



from pathlib import Path

from cg.apps.cgstats.stats import StatsAPI
from cg.meta.demultiplex.demux_post_processing import DemuxPostProcessingAPI
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flowcell import Flowcell


def test_rename_demux_result(
    demultiplexed_flowcell_working_directory: Path,
    demultiplex_context: CGConfig,
    flowcell_object: Flowcell,
    caplog,
):

    # GIVEN that this is the location of the demultiplex api
    demultiplex_context.demultiplex_api_.out_dir = demultiplexed_flowcell_working_directory
    post_demux_api: DemuxPostProcessingAPI = DemuxPostProcessingAPI(config=demultiplex_context)
    demux_dir: Path = demultiplexed_flowcell_working_directory
    demux_results: DemuxResults = DemuxResults(
        demux_dir=demux_dir, flowcell=flowcell_object, bcl_converter="bcl2fastq"
    )
    # GIVEN that there are no projects with the correct file name
    assert len(list(demux_results.projects)) == 0
    # GIVEN the location of a demultiplexed flowcell that is not finished
    assert len(list(demux_results.raw_projects)) != 0

    # WHEN renaming the files
    post_demux_api.rename_files(demux_results=demux_results)

    # THEN assert that that the files have been renamed
    assert len(list(demux_results.projects)) != 0
    assert len(list(demux_results.raw_projects)) == 0



"""Test how the api handles bundles"""
from datetime import datetime


def test_new_bundle(housekeeper_api, small_helpers):
    """Test to create a bundle with the housekeeper api"""
    # GIVEN a housekeeper api without bundles
    assert small_helpers.length_of_iterable(housekeeper_api.bundles()) == 0
    # GIVEN some bundle information
    bundle_name = "a bundle"
    created = datetime.now()
    # WHEN creating a new bundle
    bundle_obj = housekeeper_api.new_bundle(name=bundle_name, created_at=created)

    # THEN assert that the bundle object got the correct name
    assert bundle_obj.name == bundle_name
    # THEN assert that no bundle was added to the database
    assert small_helpers.length_of_iterable(housekeeper_api.bundles()) == 0


def test_add_bundle(housekeeper_api, bed_file, small_helpers):
    """Test to add a housekeeper bundle"""
    # GIVEN a empty housekeeper api
    assert small_helpers.length_of_iterable(housekeeper_api.bundles()) == 0
    # GIVEN some bundle information
    bundle_name = "a bundle"
    created = datetime.now()
    bundle_data = {
        "name": bundle_name,
        "created": created,
        "files": [{"path": bed_file, "archive": False, "tags": ["bed"]}],
    }
    # WHEN adding the bundle
    bundle_obj, _ = housekeeper_api.add_bundle(bundle_data)

    # THEN assert that no bundle was added to the database
    assert small_helpers.length_of_iterable(housekeeper_api.bundles()) == 0
    # THEN assert that the bundle object got the correct name
    assert bundle_obj.name == bundle_name


def test_add_bundle_and_commit(housekeeper_api, minimal_bundle_obj, small_helpers):
    """Test to add a housekeeper bundle and make it persistent"""
    # GIVEN a empty housekeeper api and a bundle object
    assert small_helpers.length_of_iterable(housekeeper_api.bundles()) == 0

    # WHEN commiting the bundle
    housekeeper_api.add_commit(minimal_bundle_obj)

    # THEN assert that a bundle was added to the database
    assert small_helpers.length_of_iterable(housekeeper_api.bundles()) == 1


def test_get_bundle(housekeeper_api, minimal_bundle_obj):
    """Test to add a housekeeper bundle and fetch it back"""
    # GIVEN a housekeeper api with a bundle object
    bundle_name = minimal_bundle_obj.name
    assert housekeeper_api.bundle(bundle_name) is None
    housekeeper_api.add_commit(minimal_bundle_obj)

    # WHEN fetching the bundle
    bundle_obj = housekeeper_api.bundle(bundle_name)

    # THEN assert that a bundle was fetched
    assert bundle_obj
    assert bundle_obj.name == bundle_name


def test_create_bundle_and_version(housekeeper_api, minimal_bundle_obj, small_helpers):
    """Test to create a housekeeper bundle and version"""

    # GIVEN a housekeeper api without bundles
    assert small_helpers.length_of_iterable(housekeeper_api.bundles()) == 0

    # GIVEN a bundle name
    bundle_name = "testname"

    # WHEN creating a new bundle and version
    bundle_obj = housekeeper_api.create_new_bundle_and_version(name=bundle_name)

    # THEN a new bundle with the correct name is created, with a version, and added to the database
    assert housekeeper_api.bundle(bundle_name) is not None
    assert bundle_obj.name == bundle_name
    assert bundle_obj.versions is not None
    assert small_helpers.length_of_iterable(housekeeper_api.bundles()) == 1



import logging

import click
from cg.apps.loqus import LoqusdbAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.constants.constants import DRY_RUN, SKIP_CONFIRMATION

LOG = logging.getLogger(__name__)


@click.command()
@click.option("-c", "--case_id", required=True, help="internal case id")
@DRY_RUN
@SKIP_CONFIRMATION
@click.pass_obj
def observations(context: CGConfig, case_id: str, dry_run: bool, yes: bool):
    """Delete a case from LoqusDB and reset loqusdb_id in statusDB."""

    status_db: Store = context.status_db
    loqusdb_api: LoqusdbAPI = context.loqusdb_api

    if not status_db.family(case_id):
        LOG.error("Case %s could not be found in StatusDB!", case_id)
        raise click.Abort

    if not loqusdb_api.case_exists(case_id):
        LOG.error("Case %s could not be found in LoqusDB!", case_id)
        raise click.Abort

    if dry_run:
        LOG.info("Dry run: this would delete all variants in LoqusDB for case: %s", case_id)
        return

    LOG.info("This will delete all variants in LoqusDB for case: %s", case_id)

    if yes or click.confirm("Do you want to continue?", abort=True):
        LOG.info("Deleting variants from LoqusDB for case: %s", case_id)
        loqusdb_api.delete_case(case_id=case_id)
        status_db.reset_loqusdb_observation_ids(case_id)
        status_db.commit()
        LOG.info("Removed observations for case: %s", case_id)



from cg.apps.cgstats.stats import StatsAPI

from cg.apps.cgstats.crud.delete import delete_flowcell
from cg.apps.cgstats.crud.find import get_flowcell_id


def test_delete_flowcell(populated_stats_api: StatsAPI, flowcell_name: str):
    """Test to delete a flowcell from cg-stats"""

    # GIVEN a populated StatsAPI, with an existing flow cell
    flow_cell_id = get_flowcell_id(flowcell_name=flowcell_name)
    assert flow_cell_id

    # WHEN deleting a flow cell from the StatsAPI
    delete_flowcell(manager=populated_stats_api, flowcell_name=flowcell_name)

    # THEN the flowcell should not exist any longer
    results = populated_stats_api.Flowcell.query.filter_by(flowcell_id=flow_cell_id).all()

    assert not results



"""Handler to delete data objects"""

from cg.store import models
from cg.store.api.base import BaseHandler


class DeleteDataHandler(BaseHandler):
    """Contains methods to delete business data model instances"""

    def delete_flowcell(self, flowcell_name: str):
        flowcell: models.Flowcell = self.Flowcell.query.filter(
            models.Flowcell.name == flowcell_name
        ).first()
        if flowcell:
            flowcell.delete()
            flowcell.flush()
            self.commit()



import datetime as dt
import logging
from pathlib import Path
from typing import List, Optional

from cg.apps.cgstats.db.models import Version
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.constants import HK_FASTQ_TAGS
from cg.meta.meta import MetaAPI
from cg.meta.rsync.sbatch import ERROR_RSYNC_FUNCTION, RSYNC_CONTENTS_COMMAND
from cg.models.cg_config import CGConfig
from cg.models.slurm.sbatch import Sbatch
from cg.store import models
from cg.utils.checksum.checksum import check_md5sum, extract_md5sum

LOG = logging.getLogger(__name__)


class ExternalDataAPI(MetaAPI):
    def __init__(self, config: CGConfig):
        super().__init__(config)
        self.destination_path: str = config.external.hasta
        self.source_path: str = config.external.caesar
        self.base_path: str = config.data_delivery.base_path
        self.account: str = config.data_delivery.account
        self.mail_user: str = config.data_delivery.mail_user
        self.slurm_api: SlurmAPI = SlurmAPI()
        self.RSYNC_FILE_POSTFIX: str = "_rsync_external_data"

    def create_log_dir(self, dry_run: bool, ticket: str) -> Path:
        """Creates a directory for log file to be stored"""
        timestamp: dt.datetime = dt.datetime.now()
        timestamp_str: str = timestamp.strftime("%y%m%d_%H_%M_%S_%f")
        folder_name: Path = Path("_".join([ticket, timestamp_str]))
        log_dir: Path = Path(self.base_path, folder_name)
        LOG.info("Creating folder: %s", log_dir)
        if dry_run:
            LOG.info("Would have created path %s, but this is a dry run", log_dir)
            return log_dir
        log_dir.mkdir(parents=True, exist_ok=False)
        return log_dir

    def get_source_path(
        self,
        customer: str,
        ticket: str,
        cust_sample_id: Optional[str] = "",
    ) -> Path:
        """Returns the path to where the sample files are fetched from"""
        return Path(self.source_path % customer, ticket, cust_sample_id)

    def get_destination_path(self, customer: str, lims_sample_id: Optional[str] = "") -> Path:
        """Returns the path to where the files are to be transferred"""
        return Path(self.destination_path % customer, lims_sample_id)

    def transfer_sample_files_from_source(self, dry_run: bool, ticket: str) -> None:
        """Transfers all sample files, related to given ticket, from source to destination"""
        cust: str = self.status_db.get_customer_id_from_ticket(ticket=ticket)
        log_dir: Path = self.create_log_dir(ticket=ticket, dry_run=dry_run)
        error_function: str = ERROR_RSYNC_FUNCTION.format()
        Path(self.destination_path % cust).mkdir(exist_ok=True)

        command: str = RSYNC_CONTENTS_COMMAND.format(
            source_path=self.get_source_path(customer=cust, ticket=ticket),
            destination_path=self.get_destination_path(customer=cust),
        )
        sbatch_parameters: Sbatch = Sbatch(
            job_name=ticket + self.RSYNC_FILE_POSTFIX,
            account=self.account,
            number_tasks=1,
            memory=1,
            log_dir=str(log_dir),
            email=self.mail_user,
            hours=24,
            commands=command,
            error=error_function,
        )
        self.slurm_api.set_dry_run(dry_run=dry_run)
        sbatch_content: str = self.slurm_api.generate_sbatch_content(
            sbatch_parameters=sbatch_parameters
        )
        sbatch_path: Path = Path(log_dir, ticket + self.RSYNC_FILE_POSTFIX + ".sh")
        self.slurm_api.submit_sbatch(sbatch_content=sbatch_content, sbatch_path=sbatch_path)
        LOG.info(
            "The folder {src_path} is now being rsynced to hasta".format(
                src_path=self.get_source_path(customer=cust, ticket=ticket)
            )
        )

    def get_all_fastq(self, sample_folder: Path) -> List[Path]:
        """Returns a list of all fastq.gz files in given folder"""
        all_fastqs: List[Path] = []
        for leaf in sample_folder.glob("*fastq.gz"):
            abs_path: Path = sample_folder.joinpath(leaf)
            LOG.info("Found file %s inside folder %s" % (str(abs_path), sample_folder))
            all_fastqs.append(abs_path)
        return all_fastqs

    def get_all_paths(self, customer: str, lims_sample_id: str) -> List[Path]:
        """Returns the paths of all fastq files associated to the sample"""
        fastq_folder: Path = self.get_destination_path(
            lims_sample_id=lims_sample_id, customer=customer
        )
        all_fastq_in_folder: List[Path] = self.get_all_fastq(sample_folder=fastq_folder)
        return all_fastq_in_folder

    def check_fastq_md5sum(self, fastq_path) -> Optional[Path]:
        """Returns the path of the input file if it does not match its md5sum"""
        if Path(str(fastq_path) + ".md5").exists():
            given_md5sum: str = extract_md5sum(md5sum_file=Path(str(fastq_path) + ".md5"))
            if not check_md5sum(file_path=fastq_path, md5sum=given_md5sum):
                return fastq_path

    def get_available_samples(self, folder: Path, ticket: str) -> List[models.Sample]:
        """Returns the samples from given ticket that are present in the provided folder"""
        available_folders: List[str] = [sample_path.parts[-1] for sample_path in folder.iterdir()]
        available_samples: List[models.Sample] = [
            sample
            for sample in self.status_db.get_samples_from_ticket(ticket=ticket)
            if sample.internal_id in available_folders or sample.name in available_folders
        ]
        return available_samples

    def add_files_to_bundles(
        self, fastq_paths: List[Path], last_version: Version, lims_sample_id: str
    ):
        """Adds the given fastq files to the the hk-bundle"""
        for path in fastq_paths:
            LOG.info("Adding path %s to bundle %s in housekeeper" % (path, lims_sample_id))
            self.housekeeper_api.add_file(path=path, version_obj=last_version, tags=HK_FASTQ_TAGS)

    def get_failed_fastq_paths(self, fastq_paths_to_add: List[Path]) -> List[Path]:
        failed_sum_paths: List[Path] = []
        for path in fastq_paths_to_add:
            failed_path: Optional[Path] = self.check_fastq_md5sum(path)
            if failed_path:
                failed_sum_paths.append(failed_path)
        return failed_sum_paths

    def get_fastq_paths_to_add(
        self, customer: str, hk_version: Version, lims_sample_id: str
    ) -> List[Path]:
        paths: List[Path] = self.get_all_paths(lims_sample_id=lims_sample_id, customer=customer)
        fastq_paths_to_add: List[Path] = self.housekeeper_api.check_bundle_files(
            file_paths=paths,
            bundle_name=lims_sample_id,
            last_version=hk_version,
            tags=HK_FASTQ_TAGS,
        )
        return fastq_paths_to_add

    def curate_sample_folder(self, cust_name: str, force: bool, sample_folder: Path) -> None:
        """Changes the name of the folder to the internal_id. If force is true replaces any previous folder"""
        customer: models.Customer = self.status_db.customer(internal_id=cust_name)
        customer_folder: Path = sample_folder.parent
        sample: models.Sample = self.status_db.find_samples(
            customer=customer, name=sample_folder.name
        ).first()
        if (sample and not customer_folder.joinpath(sample.internal_id).exists()) or (
            sample and force
        ):
            sample_folder.rename(customer_folder.joinpath(sample.internal_id))
        elif not sample and not self.status_db.sample(sample_folder.name):
            raise Exception(
                f"{sample_folder} is not a sample present in statusdb. Move or remove it to continue"
            )

    def add_transfer_to_housekeeper(
        self, ticket: str, dry_run: bool = False, force: bool = False
    ) -> None:
        """Creates sample bundles in housekeeper and adds the available files corresponding to the ticket to the
        bundle"""
        failed_paths: List[Path] = []
        cust: str = self.status_db.get_customer_id_from_ticket(ticket=ticket)
        destination_folder_path: Path = self.get_destination_path(customer=cust)
        for sample_folder in destination_folder_path.iterdir():
            self.curate_sample_folder(cust_name=cust, sample_folder=sample_folder, force=force)
        available_samples: List[models.Sample] = self.get_available_samples(
            folder=destination_folder_path, ticket=ticket
        )
        cases_to_start: List[dict] = []
        for sample in available_samples:
            cases_to_start.extend(
                self.status_db.cases(sample_id=sample.internal_id, exclude_analysed=True)
            )
            last_version: Version = self.housekeeper_api.get_create_version(
                bundle=sample.internal_id
            )
            fastq_paths_to_add: List[Path] = self.get_fastq_paths_to_add(
                customer=cust, hk_version=last_version, lims_sample_id=sample.internal_id
            )
            self.add_files_to_bundles(
                fastq_paths=fastq_paths_to_add,
                last_version=last_version,
                lims_sample_id=sample.internal_id,
            )
            failed_paths.extend(self.get_failed_fastq_paths(fastq_paths_to_add=fastq_paths_to_add))
        if failed_paths:
            LOG.info(
                "The following samples did not match the given md5sum: "
                + ", ".join([str(path) for path in failed_paths])
            )
            LOG.info("Changes in housekeeper will not be committed and no cases will be added")
            return
        if dry_run:
            LOG.info("No changes will be committed since this is a dry-run")
            return
        self.housekeeper_api.commit()
        for case in cases_to_start:
            self.status_db.set_case_action(case_id=case["internal_id"], action="analyze")



"""Tests for the file handlers"""
import logging
from typing import Optional

from cg.meta.upload.scout.balsamic_config_builder import BalsamicConfigBuilder
from cg.meta.upload.scout.hk_tags import CaseTags, SampleTags
from cg.meta.upload.scout.mip_config_builder import MipConfigBuilder
from cg.models.scout.scout_load_config import (
    BalsamicLoadConfig,
    MipLoadConfig,
    ScoutLoadConfig,
    ScoutMipIndividual,
)
from cg.store import models
from housekeeper.store import models as hk_models
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.madeline import MockMadelineAPI
from tests.mocks.mip_analysis_mock import MockMipAnalysis
from tests.store_helpers import StoreHelpers


def test_mip_config_builder(
    hk_version_obj: hk_models.Version,
    mip_dna_analysis_obj: models.Analysis,
    lims_api: MockLimsAPI,
    mip_analysis_api: MockMipAnalysis,
    madeline_api: MockMadelineAPI,
):
    # GIVEN a mip file handler

    # WHEN instantiating
    config_builder = MipConfigBuilder(
        hk_version_obj=hk_version_obj,
        analysis_obj=mip_dna_analysis_obj,
        lims_api=lims_api,
        mip_analysis_api=mip_analysis_api,
        madeline_api=madeline_api,
    )

    # THEN assert that the correct case tags was used
    assert isinstance(config_builder.case_tags, CaseTags)


def test_balsamic_config_builder(
    hk_version_obj: hk_models.Version, balsamic_analysis_obj: models.Analysis, lims_api: MockLimsAPI
):
    # GIVEN a balsamic file handler

    # WHEN instantiating
    file_handler = BalsamicConfigBuilder(
        hk_version_obj=hk_version_obj, analysis_obj=balsamic_analysis_obj, lims_api=lims_api
    )

    # THEN assert that the correct case tags was used
    assert isinstance(file_handler.case_tags, CaseTags)


def test_include_delivery_report_mip(mip_config_builder: MipConfigBuilder):
    # GIVEN a config builder with some data

    # GIVEN a config without a delivery report
    assert mip_config_builder.load_config.delivery_report is None

    # WHEN including the delivery report
    mip_config_builder.include_delivery_report()

    # THEN assert that the delivery report was added
    assert mip_config_builder.load_config.delivery_report is not None


def test_include_synopsis(mip_config_builder: MipConfigBuilder):
    # GIVEN a config builder with some data

    # GIVEN a config without synopsis
    assert mip_config_builder.load_config.synopsis is None

    # WHEN including the synopsis
    mip_config_builder.build_load_config()

    # THEN assert that the synopsis was added
    assert mip_config_builder.load_config.synopsis


def test_include_phenotype_groups(mip_config_builder: MipConfigBuilder):
    # GIVEN a config builder with some data

    # GIVEN a config without a phenotype groups
    assert mip_config_builder.load_config.phenotype_groups is None

    # WHEN including the phenotype groups
    mip_config_builder.include_phenotype_groups()

    # THEN assert that the phenotype groups were added
    assert mip_config_builder.load_config.phenotype_groups is not None


def test_include_phenotype_terms(mip_config_builder: MipConfigBuilder):
    # GIVEN a config builder with some data

    # GIVEN a config without a phenotype terms
    assert mip_config_builder.load_config.phenotype_terms is None

    # WHEN including the phenotype terms
    mip_config_builder.include_phenotype_terms()

    # THEN assert that the phenotype terms were added
    assert mip_config_builder.load_config.phenotype_terms is not None


def test_include_alignment_file_individual(mip_config_builder: MipConfigBuilder, sample_id: str):
    # GIVEN a mip config builder with some information

    # WHEN building the scout load config
    mip_config_builder.build_load_config()

    # THEN assert that the alignment file was added to sample id
    file_found = False
    for sample in mip_config_builder.load_config.samples:
        if sample.sample_id == sample_id:
            assert sample.alignment_path is not None
            file_found = True
    assert file_found


def test_include_mip_case_files(mip_config_builder: MipConfigBuilder):
    # GIVEN a housekeeper version bundle with some mip analysis files
    # GIVEN a case load object
    # GIVEN a mip file handler

    # WHEN including the case level files
    mip_config_builder.build_load_config()

    # THEN assert that the mandatory snv vcf was added
    assert mip_config_builder.load_config.vcf_snv


def test_include_mip_sample_files(mip_config_builder: MipConfigBuilder, sample_id: str):
    # GIVEN a housekeeper version bundle with some mip analysis files
    # GIVEN a case load object
    # GIVEN that there are no sample level mt_bam

    # GIVEN a mip file handler

    # WHEN including the case level files
    mip_config_builder.build_load_config()

    # THEN assert that the mandatory snv vcf was added
    file_found = False
    for sample in mip_config_builder.load_config.samples:
        if sample.sample_id == sample_id:
            assert sample.mt_bam is not None
            file_found = True
    assert file_found


def test_include_mip_sample_subject_id(
    mip_config_builder: MipConfigBuilder, sample_id: str, caplog
):
    # GIVEN subject_id on the sample
    caplog.set_level(level=logging.DEBUG)

    # WHEN building the config
    mip_config_builder.build_load_config()

    # THEN the subject_id was added to the scout sample
    subject_id_found = False
    for sample in mip_config_builder.load_config.samples:
        if sample.sample_id == sample_id:
            subject_id_found = True
            assert sample.subject_id is not None
    assert subject_id_found


def test_include_balsamic_case_files(balsamic_config_builder: BalsamicConfigBuilder):
    # GIVEN a housekeeper version bundle with some balsamic analysis files
    # GIVEN a case load object

    # WHEN including the case level files
    balsamic_config_builder.build_load_config()

    # THEN assert that the mandatory snv vcf was added
    assert balsamic_config_builder.load_config.vcf_cancer


def test_include_balsamic_delivery_report(balsamic_config_builder: BalsamicConfigBuilder):
    # GIVEN a housekeeper version bundle with some balsamic analysis files
    # GIVEN a case load object

    # WHEN including the case level files
    balsamic_config_builder.build_load_config()

    # THEN assert that the delivery_report exists
    assert balsamic_config_builder.load_config.delivery_report


def test_extract_generic_filepath(mip_config_builder: MipConfigBuilder):
    """Test that parsing of file path"""

    # GIVEN files paths ending with
    file_path1 = "/some/path/gatkcomb_rhocall_vt_af_chromograph_sites_X.png"
    file_path2 = "/some/path/gatkcomb_rhocall_vt_af_chromograph_sites_12.png"

    # THEN calling extracting the generic path will remove numeric id and fuffix
    generic_path = "/some/path/gatkcomb_rhocall_vt_af_chromograph_sites_"

    # THEN
    assert mip_config_builder.extract_generic_filepath(file_path1) == generic_path
    assert mip_config_builder.extract_generic_filepath(file_path2) == generic_path



"""API to prepare fastq files for analysis"""

import logging
from pathlib import Path
from typing import Dict, List

from cg.apps.crunchy import CrunchyAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.meta.compress import files
from cg.meta.compress.compress import CompressAPI
from cg.models import CompressionData
from cg.store import Store, models
from housekeeper.store import models as hk_models

LOG = logging.getLogger(__name__)


class PrepareFastqAPI:
    def __init__(self, store: Store, compress_api: CompressAPI):
        self.store: Store = store
        self.hk_api: HousekeeperAPI = compress_api.hk_api
        self.compress_api: CompressAPI = compress_api
        self.crunchy_api: CrunchyAPI = compress_api.crunchy_api

    def get_compression_objects(self, case_id: str) -> List[CompressionData]:
        """Return a list of compression objects"""
        case_obj: models.Family = self.store.family(case_id)
        compression_objects = []
        for link in case_obj.links:
            sample_id = link.sample.internal_id
            version_obj = self.compress_api.get_latest_version(sample_id)
            compression_objects.extend(files.get_spring_paths(version_obj))
        return compression_objects

    def is_spring_decompression_needed(self, case_id: str) -> bool:
        """Check if spring decompression needs to be started"""
        compression_objects = self.get_compression_objects(case_id=case_id)
        return any(
            not self.crunchy_api.is_compression_pending(compression_object)
            and not compression_object.pair_exists()
            for compression_object in compression_objects
        )

    def is_spring_decompression_running(self, case_id: str) -> bool:
        """Check if case is being decompressed"""
        compression_objects = self.get_compression_objects(case_id=case_id)
        return any(
            self.crunchy_api.is_compression_pending(compression_object)
            for compression_object in compression_objects
        )

    def can_at_least_one_sample_be_decompressed(self, case_id: str) -> bool:
        """Returns True if at least one sample can be decompressed, otherwise False"""
        compression_objects: List[CompressionData] = self.get_compression_objects(case_id=case_id)
        return any(
            self.crunchy_api.is_spring_decompression_possible(compression_object)
            for compression_object in compression_objects
        )

    def check_fastq_links(self, case_id: str) -> None:
        """Check if all fastq files are linked in housekeeper"""
        case_obj: models.Family = self.store.family(case_id)
        for link in case_obj.links:
            sample_id = link.sample.internal_id
            version_obj: hk_models.Version = self.compress_api.get_latest_version(sample_id)
            fastq_files: Dict[Path, hk_models.File] = files.get_hk_files_dict(
                tags=["fastq"], version_obj=version_obj
            )
            compression_objs: List[CompressionData] = files.get_spring_paths(version_obj)
            for compression_obj in compression_objs:
                result = True
                if compression_obj.fastq_first not in fastq_files:
                    LOG.info(
                        "Adding %s to sample %s in housekeeper"
                        % (compression_obj.fastq_first, sample_id)
                    )
                    result: bool = self.compress_api.add_decompressed_fastq(sample_obj=link.sample)
                else:
                    LOG.info(
                        "%s from sample %s is already in housekeeper"
                        % (compression_obj.fastq_first, sample_id)
                    )
                if compression_obj.fastq_second not in fastq_files:
                    LOG.info(
                        "Adding %s to sample %s in housekeeper"
                        % (compression_obj.fastq_first, sample_id)
                    )
                    result: bool = self.compress_api.add_decompressed_fastq(sample_obj=link.sample)
                else:
                    LOG.info(
                        "%s from sample %s is already in housekeeper"
                        % (compression_obj.fastq_first, sample_id)
                    )
                if not result:
                    LOG.warning("Files where not added to fastq!")



from pathlib import Path
from typing import Optional

from pydantic import BaseModel, validator

from cg.constants.loqus_upload import ObservationAnalysisTag


def check_observation_file_from_hk(file_tag: str, file: Path) -> bool:
    """Check if file exists in HK and file system"""
    if file is None:
        raise FileNotFoundError(f"No file with {file_tag} tag in housekeeper")
    if not file.exists():
        raise FileNotFoundError(f"{file} does not exist")
    return True


class ObservationsInputFiles(BaseModel):
    """Model for validating input files to Loqus"""

    case_id: str
    pedigree: Path
    snv_gbcf: Path
    snv_vcf: Path
    sv_vcf: Optional[Path] = None

    @validator("pedigree", always=True)
    def check_pedigree(cls, value) -> Path:
        """Check pedigree file exists in HK and file system"""
        if check_observation_file_from_hk(file_tag=ObservationAnalysisTag.PEDIGREE, file=value):
            return value

    @validator("snv_gbcf", always=True)
    def check_snv_gbcf(cls, value) -> Path:
        """Check snv_gbcf file exists in HK and file system"""
        if check_observation_file_from_hk(
            file_tag=ObservationAnalysisTag.CHECK_PROFILE_GBCF, file=value
        ):
            return value

    @validator("snv_vcf", always=True)
    def check_snv_vcf(cls, value) -> Path:
        """Check snv_vcf file exists in HK and file system"""
        if check_observation_file_from_hk(file_tag=ObservationAnalysisTag.SNV_VARIANTS, file=value):
            return value

    @validator("sv_vcf", always=True)
    def check_sv_vcf(cls, value) -> Path:
        """Check sv_vcf file exists in HK and file system"""
        if not value:
            return value
        if check_observation_file_from_hk(file_tag=ObservationAnalysisTag.SV_VARIANTS, file=value):
            return value



import datetime as dt

import pytest
from cg.apps.cgstats.db import models as stats_models
from cg.apps.cgstats.stats import StatsAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.meta.transfer import TransferLims
from cg.meta.transfer.flowcell import TransferFlowcell
from cg.store import Store, models
from pathlib import Path


@pytest.fixture(name="data")
def fixture_data() -> dict:
    return {
        "samples": [
            {
                "name": "ADM1136A3",
                "index": "ACGTACAT",
                "flowcell": "HJKMYBCXX",
                "type": "hiseqx",
            }
        ]
    }


@pytest.fixture(scope="function")
def store_stats() -> StatsAPI:
    """Setup base CGStats store."""
    _store = StatsAPI({"cgstats": {"database": "sqlite://", "root": "tests/fixtures/DEMUX"}})
    _store.create_all()
    yield _store
    _store.drop_all()


@pytest.fixture(scope="function")
def base_store_stats(store_stats: StatsAPI, data: dict) -> StatsAPI:
    """Setup CGStats store with sample data."""
    demuxes = {}
    for sample_data in data["samples"]:
        project: stats_models.Project = store_stats.Project(
            projectname="test", time=dt.datetime.now()
        )
        sample: stats_models.Sample = store_stats.Sample(
            samplename=sample_data["name"],
            barcode=sample_data["index"],
            limsid=sample_data["name"],
        )
        sample.project = project
        unaligned: stats_models.Unaligned = store_stats.Unaligned(
            readcounts=300000000, q30_bases_pct=85
        )
        unaligned.sample = sample

        if sample_data["flowcell"] in demuxes:
            demux = demuxes[sample_data["flowcell"]]
        else:
            flowcell: stats_models.Flowcell = store_stats.Flowcell(
                flowcellname=sample_data["flowcell"],
                flowcell_pos="A",
                hiseqtype=sample_data["type"],
                time=dt.datetime.now(),
            )
            supportparams: stats_models.Supportparams = store_stats.Supportparams(
                document_path="NA", idstring="NA"
            )
            datasource: stats_models.Datasource = store_stats.Datasource(
                document_path="NA", document_type="html"
            )
            datasource.supportparams = supportparams
            demux = store_stats.Demux()
            demux.flowcell = flowcell
            demux.datasource = datasource
            demuxes[sample_data["flowcell"]] = demux

        unaligned.demux = demux
        store_stats.add(unaligned)
    store_stats.commit()
    yield store_stats


@pytest.fixture(scope="function")
def flowcell_store(base_store: Store, data: dict) -> Store:
    """Setup store with sample data for testing flowcell transfer."""
    for sample_data in data["samples"]:
        customer_obj: models.Customer = base_store.customers().first()
        application_version: models.ApplicationVersion = base_store.application(
            "WGSPCFC030"
        ).versions[0]
        sample: models.Sample = base_store.add_sample(
            name="NA", sex="male", internal_id=sample_data["name"]
        )
        sample.customer = customer_obj
        sample.application_version = application_version
        sample.received_at = dt.datetime.now()
        base_store.add(sample)
    base_store.commit()
    yield base_store


@pytest.fixture(scope="function")
def transfer_flowcell_api(
    flowcell_store: Store, housekeeper_api: HousekeeperAPI, base_store_stats: StatsAPI
) -> TransferFlowcell:
    """Setup flowcell transfer API."""
    transfer_api = TransferFlowcell(
        db=flowcell_store, stats_api=base_store_stats, hk_api=housekeeper_api
    )
    yield transfer_api


@pytest.fixture(scope="function")
def transfer_lims_api(sample_store: Store) -> TransferLims:
    """Setup flowcell transfer API"""
    yield TransferLims(sample_store, MockLims(config=""))


class MockLims(LimsAPI):
    def __init__(self, config):
        pass

    _received_at = None
    _delivered_at = None
    _prepared_date = None
    _samples = []

    def get_received_date(self, lims_id: str):

        received_date = None
        for sample in self._samples:
            if sample.internal_id == lims_id:
                received_date = sample.received_at
        return received_date

    def mock_set_samples(self, samples):
        self._samples = samples


@pytest.fixture(scope="function")
def lims_api():

    _lims_api = MockLims(config="")
    return _lims_api


@pytest.fixture(name="external_data_directory", scope="session")
def external_data_directory(
    tmpdir_factory, customer_id: str, cust_sample_id: str, ticket: str
) -> Path:
    """Fixture that returns a customer folder with fastq.gz files in sample-directories"""
    sample1: str = cust_sample_id + "1"
    sample2: str = cust_sample_id + "2"
    cust_folder = tmpdir_factory.mktemp(customer_id, numbered=False)
    ticket_folder = cust_folder / ticket
    ticket_folder.mkdir()
    Path(ticket_folder, sample1).mkdir(exist_ok=True, parents=True)
    Path(ticket_folder, sample2).mkdir(exist_ok=True, parents=True)
    Path(ticket_folder, sample1, sample1 + "_fastq_1.fastq.gz").touch(exist_ok=True)
    Path(ticket_folder, sample1, sample1 + "_fastq_2.fastq.gz").touch(exist_ok=True)
    Path(ticket_folder, sample1, sample1 + "_fastq_2.fastq.gz.md5").touch(exist_ok=True)
    Path(ticket_folder, sample2, sample2 + "_fastq_1.fastq.gz").touch(exist_ok=True)
    Path(ticket_folder, sample2, sample2 + "_fastq_2.fastq.gz").touch(exist_ok=True)
    Path(ticket_folder, sample2, sample2 + "_fastq_2.fastq.gz.md5").touch(exist_ok=True)
    return Path(ticket_folder)



"""Gene panel specific constants"""

GENOME_BUILD_37 = "37"
GENOME_BUILD_38 = "GRCh38"

MASTER_LIST = (
    "BRAIN",
    "Cardiology",
    "CH",
    "CTD",
    "DIAB",
    "ENDO",
    "EP",
    "IBMFS",
    "IEM",
    "IF",
    "mcarta",
    "MHT",
    "MIT",
    "MOVE",
    "mtDNA",
    "NBS-M",
    "NEURODEG",
    "NMD",
    "PEDHEP",
    "PID",
    "PIDCAD",
    "OMIM-AUTO",
    "SKD",
)



"""Fixtures for crunchy api"""
import logging
from pathlib import Path
from typing import List

import pytest

from cg.constants.constants import FileFormat
from cg.io.controller import WriteFile
from cg.models import CompressionData
from cgmodels.crunchy.metadata import CrunchyMetadata
from tests.mocks.process_mock import ProcessMock

LOG = logging.getLogger(__name__)


@pytest.fixture(name="real_spring_metadata_path")
def fixture_real_spring_metadata_path(apps_dir: Path) -> Path:
    """Return the path to a SPRING metadata file"""
    return Path(apps_dir, "crunchy", "spring_metadata.json")


@pytest.fixture(name="spring_metadata")
def fixture_spring_metadata(compression_object: CompressionData) -> List[dict]:
    """Return meta data information"""
    return [
        {
            "path": compression_object.fastq_first.as_posix(),
            "file": "first_read",
            "checksum": "checksum_first_read",
            "algorithm": "sha256",
        },
        {
            "path": compression_object.fastq_second.as_posix(),
            "file": "second_read",
            "checksum": "checksum_second_read",
            "algorithm": "sha256",
        },
        {"path": compression_object.spring_path.as_posix(), "file": "spring"},
    ]


@pytest.fixture(name="crunchy_metadata_object")
def fixture_crunchy_metadata_object(spring_metadata: List[dict]) -> CrunchyMetadata:
    """Return the parsed metadata"""
    return CrunchyMetadata(files=spring_metadata)


@pytest.fixture(name="spring_metadata_file")
def fixture_spring_metadata_file(
    compression_object: CompressionData, spring_metadata: List[dict]
) -> Path:
    """Return the path to a populated SPRING metadata file"""
    metadata_path = compression_object.spring_metadata_path

    WriteFile.write_file_from_content(
        content=spring_metadata, file_format=FileFormat.JSON, file_path=metadata_path
    )
    return metadata_path


@pytest.fixture(scope="function", name="sbatch_process")
def fixture_sbatch_process(sbatch_job_number: int) -> ProcessMock:
    """Return a mocked process object"""
    slurm_process = ProcessMock(binary="sbatch")
    slurm_process.set_stdout(text=str(sbatch_job_number))
    return slurm_process


@pytest.fixture(scope="function", name="fastq_first_file")
def fixture_fastq_first_file(fastq_first_path):
    """Creates an existing FASTQ path"""
    fastq_first_path.touch()
    return fastq_first_path


@pytest.fixture(scope="function", name="fastq_second_file")
def fixture_fastq_second_file(fastq_second_path):
    """Creates an existing FASTQ path"""
    fastq_second_path.touch()
    return fastq_second_path


@pytest.fixture(scope="function", name="spring_file")
def fixture_spring_file(spring_path):
    """Creates an existing SPRING file"""
    spring_path.touch()
    return spring_path


@pytest.fixture(scope="function", name="fastq_paths")
def fixture_fastq_paths(fastq_first_path, fastq_second_path):
    """Creates FASTQ paths"""
    return {
        "fastq_first_path": fastq_first_path,
        "fastq_second_path": fastq_second_path,
    }


@pytest.fixture(scope="function", name="existing_fastq_paths")
def fixture_existing_fastq_paths(fastq_first_file, fastq_second_file):
    """Creates existing FASTQ paths"""
    return {
        "fastq_first_path": fastq_first_file,
        "fastq_second_path": fastq_second_file,
    }


@pytest.fixture(scope="function", name="compressed_fastqs")
def fixture_compressed_fastqs(existing_fastq_paths, spring_file, spring_metadata_file):
    """Creates FASTQs with corresponding SPRING and FLAG"""
    fastq_paths = existing_fastq_paths
    flag_path = spring_metadata_file

    assert spring_file.exists()
    assert flag_path.exists()
    return fastq_paths



"""Tests for reset part of the store API"""
from cg.store import Store


def test_reset_loqusdb_observation_ids(store: Store, helpers):
    # GIVEN a store with a case with loqus-links
    case = helpers.add_case(store)
    sample = helpers.add_sample(store, loqusdb_id=True)
    store.relate_sample(family=case, sample=sample, status="unknown")
    assert sample.loqusdb_id is not None

    # WHEN calling reset observations
    store.reset_loqusdb_observation_ids(case_id=case.internal_id)

    # THEN the links to observations in loqusdb should have been reset
    assert sample.loqusdb_id is None



"""
    Tests for loqusdbAPI
"""
import subprocess

import pytest
import click
from cg.apps.loqus import LoqusdbAPI
from cg.constants.constants import FileFormat
from cg.exc import CaseNotFoundError
from cg.io.controller import ReadStream
from cg.models.observations.observations_input_files import ObservationsInputFiles


def test_instantiate(loqus_config: dict):

    """Test to instantiate a loqusdb api"""
    # GIVEN a loqusdb api with some configs

    # WHEN instantiating a loqusdb api
    loqusdb = LoqusdbAPI(loqus_config)

    # THEN assert that the adapter was properly instantiated
    assert loqusdb.loqusdb_config == loqus_config["loqusdb"]["config_path"]
    assert loqusdb.loqusdb_binary == loqus_config["loqusdb"]["binary_path"]

    # WHEN instantiating with analysis_type argument as 'wes'
    loqusdb = LoqusdbAPI(loqus_config, analysis_type="wes")

    # THEN assert that the adapter was properly instantiated
    assert loqusdb.loqusdb_config == loqus_config["loqusdb_wes"]["config_path"]
    assert loqusdb.loqusdb_binary == loqus_config["loqusdb_wes"]["binary_path"]


def test_get_case(case_id, loqusdbapi, loqusdb_case_output):
    """Test to get a case via the api"""
    # GIVEN a loqusdb api
    loqusdbapi.process.stdout = loqusdb_case_output.decode("utf-8").rstrip()

    # GIVEN a case id

    # WHEN fetching a case with the adapter
    case_obj = loqusdbapi.get_case(case_id)

    # THEN assert that the correct case id is returned
    assert case_obj["case_id"] == case_id


def test_get_case_non_existing(case_id, loqusdbapi):
    """Test to get a case via the api"""
    # GIVEN a loqusdb api

    # GIVEN a case id

    # WHEN loqusdb output is empty string

    # THEN CaseNotFoundError should be raised
    with pytest.raises(CaseNotFoundError):
        loqusdbapi.get_case(case_id)


def test_get_case_command_fail(case_id: str, loqusdbapi_exception):
    """Test to see if exception is raised"""
    # GIVEN a loqusdb api
    loqusdbapi = loqusdbapi_exception

    # GIVEN a case id

    # WHEN an error occurs during fetching a case with the adapter

    # THEN assert that the error is raised
    with pytest.raises(subprocess.CalledProcessError):
        loqusdbapi.get_case(case_id)


def test_get_duplicate(
    loqusdbapi, loqusdb_duplicate_output: bytes, observation_input_files_raw: dict
):
    """Test when duplicate exists"""

    # GIVEN a loqusdb api
    loqusdbapi.process.stdout = loqusdb_duplicate_output.decode("utf-8")

    # GIVEN a duplicate output
    dup_json = ReadStream.get_content_from_stream(
        file_format=FileFormat.JSON, stream=loqusdb_duplicate_output.decode("utf-8")
    )

    # GIVEN an ObservationsInputFiles object
    files = ObservationsInputFiles(**observation_input_files_raw)

    # WHEN a duplicate exists
    duplicate = loqusdbapi.get_duplicate(files.snv_vcf)

    # THEN assert that the duplicate individual is returned
    assert dup_json == duplicate


def test_get_duplicate_non_existing(loqusdbapi, observation_input_files_raw: dict):
    """Test when no duplicate exists"""

    # GIVEN a loqusdb api

    # GIVEN an ObservationsInputFiles object
    files = ObservationsInputFiles(**observation_input_files_raw)

    # WHEN duplicate does not exist
    duplicate = loqusdbapi.get_duplicate(files.snv_vcf)

    # THEN assert that an empty dict is returned
    assert duplicate == {}


def test_load(loqusdbapi, loqusdb_output, observation_input_files_raw: dict):
    """Test to load a case"""
    # GIVEN a loqusdb api

    # GIVEN an ObservationsInputFiles object
    files = ObservationsInputFiles(**observation_input_files_raw)

    # WHEN uploading a case with 15 variants to loqusdb
    loqusdbapi.process.stderr = loqusdb_output.decode("utf-8")

    data = loqusdbapi.load(
        files.case_id, files.pedigree, files.snv_vcf, files.snv_gbcf, vcf_sv_path=files.sv_vcf
    )

    # THEN assert that the number of variants is 15
    assert data["variants"] == 15


def test_repr_string(loqus_config):

    """Test __repr__ of loqusdbAPI"""

    loqusdb = LoqusdbAPI(loqus_config)

    repr_string = repr(loqusdb)

    correct_string = (
        f"LoqusdbAPI(binary={loqus_config['loqusdb']['binary_path']},"
        f"config={loqus_config['loqusdb']['config_path']})"
    )

    assert repr_string == correct_string


def test_delete_case(case_id: str, loqusdbapi: LoqusdbAPI, loqusdb_delete_stderr: bytes):
    """Test to delete an existing case via the api."""
    # GIVEN a loqusdb api and a case_id

    # WHEN deleting an existing case from Loqusdb
    loqusdbapi.process.stderr = loqusdb_delete_stderr.decode("utf-8")
    data = loqusdbapi.delete_case(case_id)

    # THEN None should be returned and no errors raised
    assert data is None


def test_delete_case_non_existing_case(
    case_id: str, loqusdbapi: LoqusdbAPI, loqusdb_delete_nonexisting_stderr: bytes
):
    """Test to delete a non existing case via the api."""
    # GIVEN a loqusdb api and a case_id

    # WHEN deleting a non existing case from Loqusdb
    loqusdbapi.process.stderr = loqusdb_delete_nonexisting_stderr.decode("utf-8")

    # THEN assert that an error is raised
    with pytest.raises(CaseNotFoundError):
        loqusdbapi.delete_case(case_id)


def test_case_exists_true(case_id: str, loqusdbapi: LoqusdbAPI, loqusdb_case_output: bytes):
    """Test if a case exists in LoqusDB for an existing case"""
    # GIVEN a loqusdb api
    loqusdbapi.process.stdout = loqusdb_case_output.decode("utf-8").rstrip()

    # GIVEN a case id

    # WHEN fetching an existing case
    is_case = loqusdbapi.case_exists(case_id)

    # THEN assert True
    assert is_case is True


def test_case_exists_false(case_id: str, loqusdbapi: LoqusdbAPI):
    """Test if a case exists in LoqusDB for a non existing case"""
    # GIVEN a loqusdb api

    # GIVEN a case id

    # WHEN fetching a non existing case
    is_case = loqusdbapi.case_exists(case_id)

    # THEN assert False
    assert is_case is False



# -*- coding: utf-8 -*-
import os

# flask
SECRET_KEY = os.environ.get("CG_SECRET_KEY") or "thisIsNotASafeKey"
TEMPLATES_AUTO_RELOAD = True

# sqlalchemy
SQLALCHEMY_DATABASE_URI = os.environ["CG_SQL_DATABASE_URI"]
SQLALCHEMY_POOL_RECYCLE = 7200
SQLALCHEMY_TRACK_MODIFICATIONS = "FLASK_DEBUG" in os.environ

# server
CG_ENABLE_ADMIN = ("FLASK_DEBUG" in os.environ) or (os.environ.get("CG_ENABLE_ADMIN") == "1")

# lims
LIMS_HOST = os.environ["LIMS_HOST"]
LIMS_USERNAME = os.environ["LIMS_USERNAME"]
LIMS_PASSWORD = os.environ["LIMS_PASSWORD"]

OSTICKET_API_KEY = os.environ.get("OSTICKET_API_KEY")
OSTICKET_DOMAIN = os.environ.get("OSTICKET_DOMAIN")
SUPPORT_SYSTEM_EMAIL = os.environ.get("SUPPORT_SYSTEM_EMAIL")
EMAIL_URI = os.environ.get("EMAIL_URI")


# oauth
GOOGLE_OAUTH_CLIENT_ID = os.environ["GOOGLE_OAUTH_CLIENT_ID"]
GOOGLE_OAUTH_CLIENT_SECRET = os.environ["GOOGLE_OAUTH_CLIENT_SECRET"]

# invoice
TOTAL_PRICE_TRESHOLD = 750000



from pathlib import Path

import pytest

from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flowcell import Flowcell


@pytest.fixture(name="flow_cell_name")
def fixture_flow_cell_name(flowcell_full_name: str) -> str:
    return flowcell_full_name.split("_")[-1][1:]


@pytest.fixture(name="flowcell_runs")
def fixture_flowcell_runs(demultiplex_fixtures: Path) -> Path:
    return demultiplex_fixtures / "flowcell-runs"


@pytest.fixture(name="flowcell_full_name")
def fixture_flowcell_full_name() -> str:
    return "201203_A00689_0200_AHVKJCDRXX"


@pytest.fixture(name="flowcell_path")
def fixture_flowcell_path(flowcell_runs: Path, flowcell_full_name: str) -> Path:
    return flowcell_runs / flowcell_full_name


@pytest.fixture(name="flowcell_object")
def fixture_flowcell_object(flowcell_path: Path) -> Flowcell:
    flow_cell = Flowcell(flowcell_path)
    flow_cell.parse_flowcell_name()
    return Flowcell(flowcell_path)


@pytest.fixture(name="demultiplexed_flowcell")
def fixture_demultiplexed_flowcell(demultiplexed_runs: Path, flowcell_full_name: str) -> Path:
    return demultiplexed_runs / flowcell_full_name


@pytest.fixture(name="bcl2fastq_demux_results")
def fixture_bcl2fastq_demux_results(
    demultiplexed_flowcell: Path, flowcell_object: Flowcell
) -> DemuxResults:
    return DemuxResults(
        demux_dir=demultiplexed_flowcell, flowcell=flowcell_object, bcl_converter="bcl2fastq"
    )


@pytest.fixture(name="dragen_flow_cell_full_name")
def fixture_dragen_flow_cell_full_name() -> str:
    return "211101_A00187_0615_AHLG5GDRXY"


@pytest.fixture(name="dragen_flow_cell_path")
def fixture_dragen_flow_cell_path(flowcell_runs: Path, dragen_flow_cell_full_name: str) -> Path:
    return flowcell_runs / dragen_flow_cell_full_name


@pytest.fixture(name="demultiplexed_dragen_flow_cell")
def fixture_demultiplexed_dragen_flow_cell(
    demultiplexed_runs: Path, dragen_flow_cell_full_name: str
) -> Path:
    return demultiplexed_runs / dragen_flow_cell_full_name


@pytest.fixture(name="dragen_flow_cell_object")
def fixture_dragen_flow_cell_object(dragen_flow_cell_path: Path) -> Flowcell:
    flow_cell = Flowcell(dragen_flow_cell_path)
    flow_cell.parse_flowcell_name()
    return Flowcell(dragen_flow_cell_path)


@pytest.fixture(name="dragen_demux_results")
def fixture_dragen_demux_results(
    demultiplexed_dragen_flow_cell: Path, dragen_flow_cell_object: Flowcell
) -> DemuxResults:
    return DemuxResults(
        demux_dir=demultiplexed_dragen_flow_cell,
        flowcell=dragen_flow_cell_object,
        bcl_converter="dragen",
    )



"""Module for building the rsync command to send files to customer inbox on the delivery server."""
import datetime as dt
import glob
import logging
from pathlib import Path
from typing import List, Dict, Iterable

from cgmodels.trailblazer.constants import AnalysisTypes
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants.constants import FileFormat
from cg.constants.priority import SlurmQos, SLURM_ACCOUNT_TO_QOS
from cg.exc import CgError
from cg.io.controller import WriteFile
from cg.meta.meta import MetaAPI
from cg.meta.rsync.sbatch import RSYNC_COMMAND, ERROR_RSYNC_FUNCTION, COVID_RSYNC
from cg.models.cg_config import CGConfig
from cg.models.slurm.sbatch import Sbatch
from cg.store import models
from cg.constants import Pipeline

LOG = logging.getLogger(__name__)


class RsyncAPI(MetaAPI):
    def __init__(self, config: CGConfig):
        super().__init__(config)
        self.delivery_path: str = config.delivery_path
        self.destination_path: str = config.data_delivery.destination_path
        self.covid_destination_path: str = config.data_delivery.covid_destination_path
        self.covid_report_path: str = config.data_delivery.covid_report_path
        self.base_path: Path = Path(config.data_delivery.base_path)
        self.account: str = config.data_delivery.account
        self.log_dir: Path = Path(config.data_delivery.base_path)
        self.mail_user: str = config.data_delivery.mail_user
        self.slurm_quality_of_service: str = SLURM_ACCOUNT_TO_QOS[self.account] or SlurmQos.LOW
        self.pipeline: str = Pipeline.RSYNC

    @property
    def trailblazer_config_path(self) -> Path:
        """Return Path to trailblazer config."""
        return self.log_dir / "slurm_job_ids.yaml"

    @property
    def rsync_processes(self) -> Iterable[Path]:
        """Yield existing rsync processes."""
        yield from self.base_path.iterdir()

    @staticmethod
    def format_covid_destination_path(covid_destination_path: str, customer_id: str) -> str:
        """Return destination path of covid report."""
        return covid_destination_path % customer_id

    @staticmethod
    def get_trailblazer_config(slurm_job_id: int) -> Dict[str, List[str]]:
        """Return dictionary of slurm job IDs."""
        return {"jobs": [str(slurm_job_id)]}

    @staticmethod
    def write_trailblazer_config(content: dict, config_path: Path) -> None:
        """Write slurm job IDs to a .YAML file used as the trailblazer config."""
        LOG.info(f"Writing slurm jobs to {config_path.as_posix()}")
        WriteFile.write_file_from_content(
            content=content, file_format=FileFormat.YAML, file_path=config_path
        )

    @staticmethod
    def process_ready_to_clean(before: dt.datetime, process: Path) -> bool:
        """Return True if analysis is old enough to be cleaned."""

        ctime: dt.datetime = dt.datetime.fromtimestamp(process.stat().st_ctime)

        return before > ctime

    @staticmethod
    def concatenate_rsync_commands(
        folder_list: List[str], source_and_destination_paths: Dict[str, Path], ticket: str
    ) -> str:
        """Concatenates the rsync commands for each folder to be transferred."""
        commands = ""
        for folder in folder_list:
            source_path: Path = source_and_destination_paths["delivery_source_path"] / folder
            destination_path: Path = Path(
                source_and_destination_paths["rsync_destination_path"], ticket
            )
            commands += RSYNC_COMMAND.format(
                source_path=source_path, destination_path=destination_path
            )
        return commands

    def set_log_dir(self, folder_prefix: str) -> None:
        if folder_prefix not in str(self.log_dir.as_posix()):
            timestamp = dt.datetime.now()
            timestamp_str = timestamp.strftime("%y%m%d_%H_%M_%S_%f")
            folder_name = Path("_".join([folder_prefix, timestamp_str]))
            LOG.info(f"Setting log dir to: {self.base_path / folder_name}")
            self.log_dir: Path = self.base_path / folder_name

    def get_all_cases_from_ticket(self, ticket: str) -> List[models.Family]:
        return self.status_db.get_cases_from_ticket(ticket=ticket).all()

    def get_source_and_destination_paths(self, ticket: str) -> Dict[str, Path]:
        cases: List[models.Family] = self.get_all_cases_from_ticket(ticket=ticket)
        source_and_destination_paths: Dict[str, Path] = {}
        if not cases:
            LOG.warning("Could not find any cases for ticket %s", ticket)
            raise CgError()
        customer_id: str = cases[0].customer.internal_id
        source_and_destination_paths["delivery_source_path"]: Path = Path(
            self.delivery_path, customer_id, "inbox", ticket
        )
        source_and_destination_paths["rsync_destination_path"]: Path = Path(
            self.destination_path, customer_id, "inbox"
        )
        return source_and_destination_paths

    def add_to_trailblazer_api(
        self, tb_api: TrailblazerAPI, slurm_job_id: int, ticket: str, dry_run: bool
    ) -> None:
        """Add rsync process to trailblazer."""
        if dry_run:
            return
        self.write_trailblazer_config(
            content=self.get_trailblazer_config(slurm_job_id),
            config_path=self.trailblazer_config_path,
        )
        tb_api.add_pending_analysis(
            case_id=ticket,
            analysis_type=AnalysisTypes.OTHER,
            config_path=self.trailblazer_config_path.as_posix(),
            out_dir=self.log_dir.as_posix(),
            slurm_quality_of_service=self.slurm_quality_of_service,
            email=self.mail_user,
            data_analysis=Pipeline.RSYNC,
        )

    def format_covid_report_path(self, case: models.Family, ticket: str) -> str:
        """Return a formatted of covid report path."""
        covid_report_options: List[str] = glob.glob(
            self.covid_report_path % (case.internal_id, ticket)
        )
        if not covid_report_options:
            LOG.error(
                f"No report file could be found with path"
                f" {self.covid_report_path % (case.internal_id, ticket)}!"
            )
            raise CgError()
        return covid_report_options[0]

    def create_log_dir(self, dry_run: bool) -> None:
        """Create log dir."""
        log_dir: Path = self.log_dir
        LOG.info("Creating folder: %s", log_dir)
        if log_dir.exists():
            LOG.warning("Could not create %s, this folder already exist", log_dir)
        elif dry_run:
            LOG.info("Would have created path %s, but this is a dry run", log_dir)
        else:
            log_dir.mkdir(parents=True, exist_ok=True)

    def get_folders_to_deliver(
        self, case_id: str, sample_files_present: bool, case_files_present: bool
    ) -> List[str]:
        """Returns a list of all the folder names depending if sample and/or case data is to be
        transferred."""
        if not sample_files_present and not case_files_present:
            raise CgError(
                "Since neither case or sample files are present, no files will be transferred"
            )
        folder_list: List[str] = []
        if sample_files_present:
            folder_list.extend(
                [
                    sample.name
                    for sample in self.status_db.get_samples_by_family_id(family_id=case_id)
                ]
            )
        if case_files_present:
            folder_list.append(self.status_db.family(case_id).name)
        return folder_list

    def slurm_rsync_single_case(
        self,
        case_id: str,
        dry_run: bool,
        sample_files_present: bool = False,
        case_files_present: bool = False,
    ) -> int:
        """Runs rsync of a single case to the delivery server, parameters depend on delivery type."""

        ticket: str = self.status_db.get_latest_ticket_from_case(case_id=case_id)
        source_and_destination_paths: Dict[str, Path] = self.get_source_and_destination_paths(
            ticket=ticket
        )
        self.set_log_dir(folder_prefix=case_id)
        self.create_log_dir(dry_run=dry_run)

        folder_list: List[str] = self.get_folders_to_deliver(
            case_id=case_id,
            sample_files_present=sample_files_present,
            case_files_present=case_files_present,
        )
        commands: str = RsyncAPI.concatenate_rsync_commands(
            folder_list=folder_list,
            source_and_destination_paths=source_and_destination_paths,
            ticket=ticket,
        )

        return self.sbatch_rsync_commands(commands=commands, job_prefix=case_id, dry_run=dry_run)

    def run_rsync_on_slurm(self, ticket: str, dry_run: bool) -> int:
        """Runs rsync of a whole ticket folder to the delivery server."""
        self.set_log_dir(folder_prefix=ticket)
        self.create_log_dir(dry_run=dry_run)
        source_and_destination_paths: Dict[str, Path] = self.get_source_and_destination_paths(
            ticket=ticket
        )
        cases: List[models.Family] = self.get_all_cases_from_ticket(ticket=ticket)
        customer_id: str = cases[0].customer.internal_id
        if cases[0].data_analysis == Pipeline.SARS_COV_2:
            LOG.info("Delivering report for SARS-COV-2 analysis")
            commands = COVID_RSYNC.format(
                source_path=source_and_destination_paths["delivery_source_path"],
                destination_path=source_and_destination_paths["rsync_destination_path"],
                covid_report_path=self.format_covid_report_path(case=cases[0], ticket=ticket),
                covid_destination_path=self.format_covid_destination_path(
                    self.covid_destination_path, customer_id=customer_id
                ),
                log_dir=self.log_dir,
            )
        else:
            commands = RSYNC_COMMAND.format(
                source_path=source_and_destination_paths["delivery_source_path"],
                destination_path=source_and_destination_paths["rsync_destination_path"],
            )
        return self.sbatch_rsync_commands(commands=commands, job_prefix=ticket, dry_run=dry_run)

    def sbatch_rsync_commands(
        self,
        commands: str,
        job_prefix: str,
        account: str = None,
        email: str = None,
        log_dir: str = None,
        hours: int = 24,
        number_tasks: int = 1,
        memory: int = 1,
        dry_run: bool = False,
    ) -> int:
        """Instantiates a slurm api and sbatches the given commands. Default parameters can be
        overridden."""
        sbatch_parameters: Sbatch = Sbatch(
            job_name="_".join([job_prefix, "rsync"]),
            account=account or self.account,
            number_tasks=number_tasks,
            memory=memory,
            log_dir=log_dir or self.log_dir.as_posix(),
            email=email or self.mail_user,
            hours=hours,
            quality_of_service=self.slurm_quality_of_service,
            commands=commands,
            error=ERROR_RSYNC_FUNCTION.format(),
            exclude="--exclude=gpu-compute-0-[0-1],cg-dragen",
        )
        slurm_api = SlurmAPI()
        slurm_api.set_dry_run(dry_run=dry_run)
        sbatch_content: str = slurm_api.generate_sbatch_content(sbatch_parameters=sbatch_parameters)
        sbatch_path = self.log_dir / "_".join([job_prefix, "rsync.sh"])
        sbatch_number: int = slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        return sbatch_number



"""Code for uploading observations data via CLI"""
import logging
from typing import Optional

import click
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.loqus import LoqusdbAPI
from cg.constants import Pipeline
from cg.exc import DuplicateRecordError, DuplicateSampleError
from cg.meta.upload.observations import UploadObservationsAPI
from cg.models.cg_config import CGConfig
from cg.store import Store

from .utils import LinkHelper

LOG = logging.getLogger(__name__)


@click.command()
@click.option("-c", "--case_id", help="internal case id, leave empty to process all")
@click.option("-l", "--case-limit", type=int, help="maximum number of cases to upload")
@click.option("--dry-run", is_flag=True, help="only print cases to be processed")
@click.pass_obj
def observations(
    context: CGConfig, case_id: Optional[str], case_limit: Optional[int], dry_run: bool
):
    """Upload observations from an analysis to LoqusDB."""

    click.echo(click.style("----------------- OBSERVATIONS ----------------"))
    status_db: Store = context.status_db
    housekeeper_api: HousekeeperAPI = context.housekeeper_api

    loqus_apis = {
        "wgs": LoqusdbAPI(context.dict()),
        "wes": LoqusdbAPI(context.dict(), analysis_type="wes"),
    }

    if case_id:
        families_to_upload = [status_db.family(case_id)]
    else:
        families_to_upload = status_db.observations_to_upload()

    nr_uploaded = 0
    for case_obj in families_to_upload:

        if case_limit is not None and nr_uploaded >= case_limit:
            LOG.info(
                "Uploaded %d cases, observations upload will now stop",
                nr_uploaded,
            )
            return

        if not case_obj.customer.loqus_upload:
            LOG.info(
                "%s: %s not whitelisted for upload to loqusdb. Skipping!",
                case_obj.internal_id,
                case_obj.customer.internal_id,
            )
            continue

        if case_obj.data_analysis.lower() != str(Pipeline.MIP_DNA):
            LOG.info("%s: has non-MIP data_analysis. Skipping!", case_obj.internal_id)
            continue

        if not LinkHelper.all_samples_are_non_tumour(case_obj.links):
            LOG.info("%s: has tumour samples. Skipping!", case_obj.internal_id)
            continue

        analysis_list = LinkHelper.get_analysis_type_for_each_link(case_obj.links)
        if len(set(analysis_list)) != 1 or analysis_list[0] not in (
            "wes",
            "wgs",
        ):
            LOG.info(
                "%s: Undetermined analysis type (wes or wgs) or mixed analyses. Skipping!",
                case_obj.internal_id,
            )
            continue

        analysis_type = analysis_list[0]

        if dry_run:
            LOG.info("%s: Would upload observations", case_obj.internal_id)
            continue

        upload_observations_api = UploadObservationsAPI(
            status_api=status_db, hk_api=housekeeper_api, loqus_api=loqus_apis[analysis_type]
        )

        try:
            upload_observations_api.process(case_obj.analyses[0])
            LOG.info("%s: observations uploaded!", case_obj.internal_id)
            nr_uploaded += 1
        except (DuplicateRecordError, DuplicateSampleError) as error:
            LOG.info(
                "%s: skipping observations upload: %s",
                case_obj.internal_id,
                error,
            )
        except FileNotFoundError as error:
            LOG.info(
                "%s: skipping observations upload: %s",
                case_obj.internal_id,
                error,
            )



"""
    Tests for VogueAPI
"""
import logging
import subprocess
from unittest import mock

from cg.apps.vogue import VogueAPI


def test_instatiate(vogue_config):

    """Test to instantiate a vogue api"""

    # GIVEN a vogue api with a config

    # WHEN instantiating a vogue api
    vogue_api = VogueAPI(vogue_config)

    # THEN assert that the adapter was properly instantiated
    assert vogue_api.vogue_binary == vogue_config["vogue"]["binary_path"]


def test_load_bioinfo_sample(vogue_config, caplog):

    """Test load bioinfo sample upload in vogue api"""

    # GIVEN a vogue api and a vogue config
    vogue_api = VogueAPI(vogue_config)
    dummy_stdout = "dummy_stdout"
    dummy_stderr = "dummy_stderr"
    dummy_returncode_success = 0
    test_load_bioinfo_sample_inputs = {"analysis_case_name": "dummy_case"}
    load_bioinfo_sample_call = [
        "load",
        "bioinfo",
        "sample",
        "--analysis-case",
        test_load_bioinfo_sample_inputs["analysis_case_name"],
    ]
    load_bioinfo_sample_call_command = " ".join(load_bioinfo_sample_call)
    caplog.set_level(logging.INFO)

    # WHEN instantiating a vogue api and input data
    with mock.patch.object(subprocess, "run") as mocked:
        mocked.return_value.stdout = dummy_stdout.encode("utf-8")
        mocked.return_value.stderr = dummy_stderr.encode("utf-8")
        mocked.return_value.returncode = dummy_returncode_success
        vogue_api.load_bioinfo_sample(load_bioinfo_inputs=test_load_bioinfo_sample_inputs)

        # THEN assert that command is in log output
        assert load_bioinfo_sample_call_command in caplog.text


def test_load_bioinfo_raw(vogue_config, caplog):

    """Test load bioinfo raw upload in vogue api"""

    # GIVEN a vogue api and a vogue config
    vogue_api = VogueAPI(vogue_config)
    dummy_stdout = "dummy_stdout"
    dummy_stderr = "dummy_stderr"
    dummy_returncode_success = 0
    test_load_bioinfo_raw_inputs = {
        "samples": "dummy_sample",
        "analysis_result_file": "dummy_file_name",
        "analysis_type": "dummy_analysis_type",
        "analysis_case_name": "dummy_analysis_case_name",
        "analysis_workflow_version": "dummy_version_3.1415",
        "case_analysis_type": "dummy_multiqc",
        "analysis_workflow_name": "dummy_best_workflow",
    }

    load_bioinfo_raw_call = [
        "load",
        "bioinfo",
        "raw",
        "--sample-list",
        test_load_bioinfo_raw_inputs["samples"],
        "--analysis-result",
        test_load_bioinfo_raw_inputs["analysis_result_file"],
        "--analysis-type",
        test_load_bioinfo_raw_inputs["analysis_type"],
        "--analysis-case",
        test_load_bioinfo_raw_inputs["analysis_case_name"],
        "--workflow-version",
        test_load_bioinfo_raw_inputs["analysis_workflow_version"],
        "--case-analysis-type",
        test_load_bioinfo_raw_inputs["case_analysis_type"],
        "--analysis-workflow",
        test_load_bioinfo_raw_inputs["analysis_workflow_name"],
    ]
    load_bioinfo_raw_call_command = " ".join(load_bioinfo_raw_call)
    caplog.set_level(logging.INFO)

    # WHEN instantiating a vogue api and input data
    with mock.patch.object(subprocess, "run") as mocked:
        mocked.return_value.stdout = dummy_stdout.encode("utf-8")
        mocked.return_value.stderr = dummy_stderr.encode("utf-8")
        mocked.return_value.returncode = dummy_returncode_success
        vogue_api.load_bioinfo_raw(load_bioinfo_inputs=test_load_bioinfo_raw_inputs)

        # THEN assert that command is in log output
        assert load_bioinfo_raw_call_command in caplog.text


def test_load_bioinfo_process(vogue_config, caplog):

    """Test load bioinfo process upload in vogue api"""

    # GIVEN a vogue api and a vogue config
    vogue_api = VogueAPI(vogue_config)
    dummy_stdout = "dummy_stdout"
    dummy_stderr = "dummy_stderr"
    dummy_returncode_success = 0
    test_load_bioinfo_process_inputs = {
        "analysis_type": "dummy_analysis_type",
        "analysis_case_name": "dummy_analysis_case_name",
        "analysis_workflow_name": "dummy_best_workflow",
        "analysis_workflow_version": "dummy_version_3.1415",
        "case_analysis_type": "dummy_multiqc",
    }

    load_bioinfo_process_call = [
        "load",
        "bioinfo",
        "process",
        "--analysis-type",
        test_load_bioinfo_process_inputs["analysis_type"],
        "--analysis-case",
        test_load_bioinfo_process_inputs["analysis_case_name"],
        "--analysis-workflow",
        test_load_bioinfo_process_inputs["analysis_workflow_name"],
        "--workflow-version",
        test_load_bioinfo_process_inputs["analysis_workflow_version"],
        "--case-analysis-type",
        test_load_bioinfo_process_inputs["case_analysis_type"],
    ]

    load_bioinfo_process_call_command = " ".join(load_bioinfo_process_call)
    caplog.set_level(logging.INFO)

    # WHEN instantiating a vogue api and input data
    with mock.patch.object(subprocess, "run") as mocked:
        mocked.return_value.stdout = dummy_stdout.encode("utf-8")
        mocked.return_value.stderr = dummy_stderr.encode("utf-8")
        mocked.return_value.returncode = dummy_returncode_success
        vogue_api.load_bioinfo_process(
            load_bioinfo_inputs=test_load_bioinfo_process_inputs,
            cleanup_flag=True,
        )

        # THEN assert that command is in log output
        assert load_bioinfo_process_call_command in caplog.text


def test_vogue_load_reagent_labels(vogue_api, caplog):
    caplog.set_level(logging.DEBUG)
    # GIVEN a vogue api

    # WHEN running vogue load reagent_label with some days as argument
    vogue_api.load_reagent_labels(days=1)

    # THEN assert vogue output is comunicated
    assert "vogue output" in caplog.text


def test_vogue_load_samples(vogue_api, caplog):
    caplog.set_level(logging.DEBUG)

    # GIVEN a vogue api

    # WHEN running vogue load samples with some days as argument
    vogue_api.load_samples(days=1)

    # THEN assert vogue output is comunicated
    assert "vogue output" in caplog.text


def test_vogue_load_flowcells(vogue_api, caplog):
    caplog.set_level(logging.DEBUG)
    # GIVEN a vogue api

    # WHEN running vogue load flowcells with some days as argument
    vogue_api.load_flowcells(days=1)

    # THEN assert vogue output is comunicated
    assert "vogue output" in caplog.text
